#!/usr/bin/env python3

# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

"""Test the ability to free memory from a toast.Cache.

This stores the following objects per detector in a Cache:

- Detector signal as float64
- Detector flags as uint8
- Detector pointing pixel numbers as int64
- Detector pointing weights as float32

It reports the memory available before and after this allocation.
Then it frees the buffers of a given type from all detectors and
compares the resulting change to what is expected.

"""
import os
import re
import sys
import argparse
import traceback
import psutil

import numpy as np

from toast.utils import Logger

from toast.cache import Cache


def main():
    log = Logger.get()

    parser = argparse.ArgumentParser(description="Allocate and free cache objects.")

    parser.add_argument(
        "--ndet", required=False, type=int, default=10, help="The number of detectors"
    )

    parser.add_argument(
        "--nobs", required=False, type=int, default=2, help="The number of observations"
    )

    parser.add_argument(
        "--obsminutes",
        required=False,
        type=int,
        default=60,
        help="The number of minutes in each observation.",
    )

    parser.add_argument(
        "--rate", required=False, type=float, default=37.0, help="The sample rate."
    )

    parser.add_argument(
        "--nloop",
        required=False,
        type=int,
        default=2,
        help="The number of allocate / free loops",
    )

    args = parser.parse_args()

    log.info("Input parameters:")
    log.info("  {} observations".format(args.nobs))
    log.info("  {} minutes per obs".format(args.obsminutes))
    log.info("  {} detectors per obs".format(args.ndet))
    log.info("  {}Hz sample rate".format(args.rate))

    nsampobs = int(args.obsminutes * 60 * args.rate)

    nsamptot = args.ndet * args.nobs * nsampobs

    log.info("{} total samples across all detectors and observations".format(nsamptot))

    bytes_sigobs = nsampobs * 8
    bytes_sigtot = nsamptot * 8
    bytes_flagobs = nsampobs * 1
    bytes_flagtot = nsamptot * 1
    bytes_pixobs = nsampobs * 8
    bytes_pixtot = nsamptot * 8
    bytes_wtobs = 3 * nsampobs * 4
    bytes_wttot = 3 * nsamptot * 4

    bytes_tot = bytes_sigtot + bytes_flagtot + bytes_pixtot + bytes_wttot
    bytes_tot_mb = bytes_tot / 2 ** 20
    log.info(
        "{} total bytes ({:0.2f}MB) of data expected".format(bytes_tot, bytes_tot_mb)
    )

    for lp in range(args.nloop):
        log.info("Allocation loop {:02d}".format(lp))
        vmem = psutil.virtual_memory()._asdict()
        avstart = vmem["available"]
        avstart_mb = avstart / 2 ** 20
        log.info("  Starting with {:0.2f}MB of available memory".format(avstart_mb))

        # The list of Caches, one per "observation"
        caches = list()

        # This structure holds external references to cache objects, to ensure that we
        # can destroy objects and free memory, even if external references are held.
        refs = list()

        for ob in range(args.nobs):
            ch = Cache()
            rf = dict()
            for det in range(args.ndet):
                dname = "{:04d}".format(det)
                cname = "{}_sig".format(dname)
                rf[cname] = ch.create(cname, np.float64, (nsampobs,))
                cname = "{}_flg".format(dname)
                rf[cname] = ch.create(cname, np.uint8, (nsampobs,))
                cname = "{}_pix".format(dname)
                rf[cname] = ch.create(cname, np.int64, (nsampobs,))
                cname = "{}_wgt".format(dname)
                rf[cname] = ch.create(cname, np.float32, (nsampobs, 3))
            caches.append(ch)
            refs.append(rf)

        vmem = psutil.virtual_memory()._asdict()
        avpost = vmem["available"]
        avpost_mb = avpost / 2 ** 20
        log.info("  After allocation, {:0.2f}MB of available memory".format(avpost_mb))

        diff = avstart_mb - avpost_mb
        diffperc = 100.0 * np.absolute(diff - bytes_tot_mb) / bytes_tot_mb
        log.info(
            "  Difference is {:0.2f}MB, expected {:0.2f}MB ({:0.2f}% residual)".format(
                diff, bytes_tot_mb, diffperc
            )
        )

        for suf in ["wgt", "pix", "flg", "sig"]:
            for ob, ch in zip(range(args.nobs), caches):
                for det in range(args.ndet):
                    dname = "{:04d}".format(det)
                    ch.destroy("{}_{}".format(dname, suf))

        vmem = psutil.virtual_memory()._asdict()
        avfinal = vmem["available"]
        avfinal_mb = avfinal / 2 ** 20
        log.info(
            "  After destruction, {:0.2f}MB of available memory".format(avfinal_mb)
        )

        diff = avfinal_mb - avpost_mb
        diffperc = 100.0 * np.absolute(diff - bytes_tot_mb) / bytes_tot_mb
        log.info(
            "  Difference is {:0.2f}MB, expected {:0.2f}MB ({:0.2f}% residual)".format(
                diff, bytes_tot_mb, diffperc
            )
        )

    return


if __name__ == "__main__":
    try:
        main()
    except:
        exc_type, exc_value, exc_traceback = sys.exc_info()
        lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
        print("".join(lines), flush=True)



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI

import numpy as np

from ..dist import distribute_samples

from ..cache import Cache

from .. import qarray as qa

from ..timing import function_timer

from .interval import Interval


class TOD(object):
    """
    Base class for an object that provides detector pointing and
    timestreams for a single observation.

    This class provides high-level functions that are common to all derived
    classes.  It also defines the internal methods that should be overridden
    by all derived classes.  These internal methods throw an exception if they
    are called.  A TOD base class should never be directly instantiated.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which the
            data is distributed, or None.
        detectors (list):  The list of detector names.
        samples (int):  The total number of samples.
        detindx (dict): the detector indices for use in simulations.  Default is
            { x[0] : x[1] for x in zip(detectors, range(len(detectors))) }.
        detranks (int):  The dimension of the process grid in the detector
            direction.  If not None, the MPI communicator size must be evenly divisible
            by this number.
        detbreaks (list):  Optional list of hard breaks in the detector
            distribution.
        sampsizes (list):  Optional list of sample chunk sizes which
            cannot be split.
        sampbreaks (list):  Optional list of hard breaks in the sample
            distribution.
        meta (dict):  Optional dictionary of metadata properties.

    """

    def __init__(
        self,
        mpicomm,
        detectors,
        samples,
        detindx=None,
        detranks=1,
        detbreaks=None,
        sampsizes=None,
        sampbreaks=None,
        meta=None,
    ):
        self._mpicomm = mpicomm
        self._detranks = detranks

        self._sampranks = 1
        self._rank_det = 0
        self._rank_samp = 0
        self._comm_row = None
        self._comm_col = None

        rank = 0

        if mpicomm is None:
            if detranks != 1:
                raise RuntimeError("MPI is disabled, so detranks must equal 1")
        else:
            rank = mpicomm.rank
            if mpicomm.size % detranks != 0:
                raise RuntimeError(
                    "The number of detranks ({}) does not divide evenly into the "
                    "communicator size ({})".format(detranks, mpicomm.size)
                )
            self._sampranks = mpicomm.size // detranks
            self._rank_det = mpicomm.rank // self._sampranks
            self._rank_samp = mpicomm.rank % self._sampranks

            # Split the main communicator into process row and column
            # communicators, since this is useful for gathering data in some
            # operations.

            if self._sampranks == 1:
                self._comm_row = MPI.COMM_SELF
            else:
                self._comm_row = self._mpicomm.Split(self._rank_det, self._rank_samp)

            if self._detranks == 1:
                self._comm_col = MPI.COMM_SELF
            else:
                self._comm_col = self._mpicomm.Split(self._rank_samp, self._rank_det)

        self._dets = detectors

        self._nsamp = samples

        self._sizes = sampsizes

        self.meta = meta
        if meta is None:
            self.meta = {}

        if detindx is not None:
            for d in self._dets:
                if d not in detindx:
                    raise RuntimeError("detindx must have a value for every detector")
            self._detindx = detindx
        else:
            self._detindx = {x[0]: x[1] for x in zip(detectors, range(len(detectors)))}

        # if sizes is specified, it must be consistent with
        # the total number of samples.
        if self._sizes is not None:
            test = np.sum(self._sizes)
            if samples != test:
                raise RuntimeError(
                    "Sum of sampsizes ({}) does not equal total samples ({})"
                    "".format(test, samples)
                )

        (self._dist_dets, self._dist_samples, self._dist_sizes) = distribute_samples(
            self._mpicomm,
            self._dets,
            self._nsamp,
            detranks=self._detranks,
            detbreaks=detbreaks,
            sampsizes=sampsizes,
            sampbreaks=sampbreaks,
        )

        if self._sizes is None:
            # in this case, the chunks just come from the uniform distribution.
            self._sizes = [self._dist_samples[x][1] for x in range(self._sampranks)]

        if rank == 0:
            # check that all processes have some data, otherwise print warning
            for d in range(self._detranks):
                if len(self._dist_dets[d]) == 0:
                    print(
                        "WARNING: detector rank {} has no detectors"
                        " assigned.".format(d)
                    )
            for r in range(self._sampranks):
                if self._dist_samples[r][1] <= 0:
                    print(
                        "WARNING: sample rank {} has no data assigned "
                        "in TOD.".format(r)
                    )

        self.cache = Cache()

    TIMESTAMP_NAME = "timestamps"
    """Default cache name for timestamps."""

    COMMON_FLAG_NAME = "common_flags"
    """Default cache name for common flags."""

    VELOCITY_NAME = "velocity"
    """Default cache name for velocity."""

    POSITION_NAME = "position"
    """Default cache name for position."""

    SIGNAL_NAME = "signal"
    """Default cache name for signal."""

    FLAG_NAME = "flags"
    """Default cache name for flags."""

    POINTING_NAME = "quat"
    """Default cache name for pointing quaternions."""

    HWP_ANGLE_NAME = "hwp_angle"
    """Default cache name for HWP angle."""

    def __repr__(self):
        clsname = self.__class__.__name__
        csize = self.cache.report(silent=True)
        lsamp = self._dist_samples[self._rank_samp]
        ldet = self._dist_dets[self._rank_det]
        ldetstr = ["      {}".format(x) for x in ldet]
        lines = [
            "  {} total detectors and {} total samples".format(
                len(self._dets), self._nsamp
            ),
            "  Using MPI communicator {}".format(self._mpicomm),
            "    In grid dimensions {} sample ranks x {} detranks".format(
                self._sampranks, self._detranks
            ),
            "  Process at ({}, {}) in grid has data for:".format(
                self._rank_samp, self._rank_det
            ),
            "    Samples {} - {} (inclusive)".format(lsamp[0], lsamp[0] + lsamp[1] - 1),
            "    Detectors:",
        ]
        lines.extend(ldetstr)
        lines.append("    Cache contains {} bytes".format(csize))
        return "<{}\n{}\n>".format(clsname, "\n".join(lines))

    @property
    def detectors(self):
        """
        (list): The total list of detectors.
        """
        return self._dets

    def detoffset(self):
        """
        Return dictionary of detector quaternions.

        This returns a dictionary with the detector names as the keys and the
        values are 4-element numpy arrays containing the quaternion offset
        from the boresight.

        Args:
            None

        Returns (dict):
            the dictionary of quaternions.
        """
        raise NotImplementedError("Fell through to TOD base class method")
        return None

    @property
    def detindx(self):
        """
        (dict): The detector indices.
        """
        return self._detindx

    @property
    def local_dets(self):
        """
        (list): The detectors assigned to this process.
        """
        return self._dist_dets[self._rank_det]

    @property
    def total_chunks(self):
        """
        (list): the full list of sample chunk sizes that were used in the
            data distribution.
        """
        return self._sizes

    @property
    def dist_chunks(self):
        """
        (list): this is a list of 2-tuples, one for each column of the process
        grid.  Each element of the list is the same as the information returned
        by the "local_chunks" member for a given process column.
        """
        return self._dist_sizes

    @property
    def local_chunks(self):
        """
        (2-tuple): the first element of the tuple is the index of the
        first chunk assigned to this process (i.e. the index in the list
        given by the "total_chunks" member).  The second element of the
        tuple is the number of chunks assigned to this process.
        """
        return self._dist_sizes[self._rank_samp]

    def local_times(self, name=None, **kwargs):
        """Timestamps covering locally stored data.

        Args:
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a timestamp vector.  If 'name' is None
            a default name 'timestamps' is used and the vector may be
            constructed and cached using the 'read_times' method.
            If 'name' is given, then the times must already be cached.

        """
        if name is None:
            cachename = self.TIMESTAMP_NAME
            if not self.cache.exists(cachename):
                times = self.read_times(**kwargs)
                self.cache.put(cachename, times)
        else:
            cachename = name
        return self.cache.reference(cachename)

    def local_signal(self, det, name=None, **kwargs):
        """Locally stored signal.

        Args:
            det (str):  Name of the detector.
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a signal vector.  If 'name' is None
            a default name 'signal' is used and the vector may be
            constructed and cached using the 'read' method.
            If 'name' is given, then the signal must already be cached.

        """
        if name is None:
            cachename = "{}_{}".format(self.SIGNAL_NAME, det)
            if not self.cache.exists(cachename):
                signal = self.read(detector=det, **kwargs)
                self.cache.put(cachename, signal)
        else:
            cachename = "{}_{}".format(name, det)
        return self.cache.reference(cachename)

    def local_pointing(self, det, name=None, **kwargs):
        """Locally stored pointing.

        Args:
            det (str):  Name of the detector.
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a pointing array.  If 'name' is None
            a default name 'quat' is used and the array may be
            constructed and cached using the 'read_pntg' method.
            If 'name' is given, then the pointing must already be cached.

        """
        if name is None:
            cachename = "{}_{}".format(self.POINTING_NAME, det)
            if not self.cache.exists(cachename):
                quats = self.read_pntg(detector=det, **kwargs)
                self.cache.put(cachename, quats)
        else:
            cachename = "{}_{}".format(name, det)
        return self.cache.reference(cachename)

    def local_position(self, name=None, **kwargs):
        """Locally stored position.

        Args:
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a position array.  If 'name' is None
            a default name 'position' is used and the array may be
            constructed and cached using the 'read_position' method.
            If 'name' is given, then the position must already be cached.

        """
        if name is None:
            cachename = self.POSITION_NAME
            if not self.cache.exists(cachename):
                pos = self.read_position(**kwargs)
                self.cache.put(cachename, pos)
        else:
            cachename = name
        return self.cache.reference(cachename)

    def local_velocity(self, name=None, **kwargs):
        """Locally stored velocity.

        Args:
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a velocity array.  If 'name' is None
            a default name 'velocity' is used and the array may be
            constructed and cached using the 'read_velocity' method.
            If 'name' is given, then the velocity must already be cached.

        """
        if name is None:
            cachename = self.VELOCITY_NAME
            if not self.cache.exists(cachename):
                vel = self.read_velocity(**kwargs)
                self.cache.put(cachename, vel)
        else:
            cachename = name
        return self.cache.reference(cachename)

    def local_flags(self, det, name=None, **kwargs):
        """Locally stored flags.

        Args:
            det (str):  Name of the detector.
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a flag vector.  If 'name' is None
            a default name 'flags' is used and the vector may be
            constructed and cached using the 'read_flags' method.
            If 'name' is given, then the flags must already be cached.

        """
        if name is None:
            cachename = "{}_{}".format(self.FLAG_NAME, det)
            if not self.cache.exists(cachename):
                flags = self.read_flags(detector=det, **kwargs)
                self.cache.put(cachename, flags)
        else:
            cachename = "{}_{}".format(name, det)
        return self.cache.reference(cachename)

    def local_common_flags(self, name=None, **kwargs):
        """Locally stored common flags.

        Args:
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a common flag vector.  If 'name' is None
            a default name 'common_flags' is used and the vector may be
            constructed and cached using the 'read_common_flags' method.
            If 'name' is given, then the flags must already be cached.

        """
        if name is None:
            cachename = self.COMMON_FLAG_NAME
            if not self.cache.exists(cachename):
                common_flags = self.read_common_flags(**kwargs)
                self.cache.put(cachename, common_flags)
        else:
            cachename = name
        return self.cache.reference(cachename)

    def local_hwp_angle(self, name=None, **kwargs):
        """Locally stored half-wave plate angle.

        Args:
            name (str):  Optional cache key to use.
        Returns:
            A cache reference to a hwp angle vector.  If 'name' is None
            a default name 'hwp_angle' is used and the vector may be
            constructed and cached using the 'read_hwp_angle' method.
            If 'name' is given, then the angles must already be cached.

        """
        if name is None:
            cachename = self.HWP_ANGLE_NAME
            if not self.cache.exists(cachename):
                hwp_angle = self.read_hwp_angle(**kwargs)
                if hwp_angle is None:
                    return None
                self.cache.put(cachename, hwp_angle)
        else:
            cachename = name
        return self.cache.reference(cachename)

    def local_intervals(self, intervals):
        """Translate observation-wide intervals into local sample indices."""
        if intervals is None:
            intervals = [
                Interval(start=0, stop=0, first=0, last=self.total_samples - 1)
            ]
        offset, nsamp = self.local_samples
        local_intervals = []
        times = self.local_times()
        if len(times) != nsamp:
            raise RuntimeError(
                "Length of cached timestamps does not match local samples. "
                "Cannot produce local intervals."
            )
        for ival in intervals:
            previous_last = None
            if ival.last >= offset and ival.first < offset + nsamp:
                local_first = max(0, ival.first - offset)
                if previous_last is not None and previous_last >= local_first:
                    raise RuntimeError("Provided intervals overlap")
                local_last = min(nsamp - 1, ival.last - offset)
                previous_last = local_last
                local_start = times[local_first]
                local_stop = times[local_last]
                local_intervals.append(
                    Interval(
                        start=local_start,
                        stop=local_stop,
                        first=local_first,
                        last=local_last,
                    )
                )
        return local_intervals

    @property
    def total_samples(self):
        """
        (int): the total number of samples in this TOD.
        """
        return self._nsamp

    @property
    def dist_samples(self):
        """
        (list): This is a list of 2-tuples, with one element per column
            of the process grid.  Each tuple is the same information
            returned by the "local_samples" member for the corresponding
            process grid column rank.
        """
        return self._dist_samples

    @property
    def local_samples(self):
        """
        (2-tuple): The first element of the tuple is the first global
            sample assigned to this process.  The second element of
            the tuple is the number of samples assigned to this
            process.
        """
        return self._dist_samples[self._rank_samp]

    @property
    def mpicomm(self):
        """
        (mpi4py.MPI.Comm): the communicator assigned to this TOD.
        """
        return self._mpicomm

    @property
    def grid_size(self):
        """
        (tuple): the dimensions of the process grid in (detector, sample)
            directions.
        """
        return (self._detranks, self._sampranks)

    @property
    def grid_ranks(self):
        """
        (tuple): the ranks of this process in the (detector, sample)
            directions.
        """
        return (self._rank_det, self._rank_samp)

    @property
    def grid_comm_row(self):
        """
        (mpi4py.MPI.Comm): a communicator across all detectors in the same
            row of the process grid (or None).
        """
        return self._comm_row

    @property
    def grid_comm_col(self):
        """
        (mpi4py.MPI.Comm): a communicator across all detectors in the same
            column of the process grid (or None).
        """
        return self._comm_col

    def _get(self, detector, start, n):
        raise NotImplementedError("Fell through to TOD._get base class method")
        return None

    def _put(self, detector, start, data):
        raise NotImplementedError("Fell through to TOD._put base class method")
        return

    def _get_boresight(self, start, n):
        raise NotImplementedError(
            "Fell through to TOD._get_boresight base class method"
        )
        return None

    def _put_boresight(self, start, data):
        raise NotImplementedError(
            "Fell through to TOD._put_boresight base class method"
        )
        return

    def _get_boresight_azel(self, start, n):
        raise NotImplementedError(
            "Fell through to TOD._get_boresight_azel base class method"
        )
        return None

    def _put_boresight_azel(self, start, data):
        raise NotImplementedError(
            "Fell through to TOD._put_boresight_azel base class method"
        )
        return

    def _get_pntg(self, detector, start, n):
        raise NotImplementedError("Fell through to TOD._get_pntg base class method")
        return None

    def _put_pntg(self, detector, start, data):
        raise NotImplementedError("Fell through to TOD._put_pntg base class method")
        return

    def _get_flags(self, detector, start, n):
        raise NotImplementedError("Fell through to TOD._get_flags base class method")
        return None

    def _put_flags(self, detector, start, flags):
        raise NotImplementedError("Fell through to TOD._put_flags base class method")
        return

    def _get_common_flags(self, start, n):
        raise NotImplementedError(
            "Fell through to TOD._get_common_flags base class method"
        )
        return None

    def _put_common_flags(self, start, flags):
        raise NotImplementedError(
            "Fell through to TOD._put_common_flags base class method"
        )
        return None

    def _get_hwp_angle(self, start, n):
        raise NotImplementedError(
            "Fell through to TOD._get_hwp_angle base class method"
        )
        return None

    def _put_hwp_angle(self, start, flags):
        raise NotImplementedError(
            "Fell through to TOD._put_hwp_angle base class method"
        )
        return None

    def _get_times(self, start, n):
        raise NotImplementedError("Fell through to TOD._get_times base class method")
        return None

    def _put_times(self, start, stamps):
        raise NotImplementedError("Fell through to TOD._put_times base class method")
        return None

    def _get_position(self, start, n):
        raise NotImplementedError("Fell through to TOD._get_position base class method")
        return None

    def _put_position(self, start, pos):
        raise NotImplementedError("Fell through to TOD._put_position base class method")
        return

    def _get_velocity(self, start, n):
        raise NotImplementedError("Fell through to TOD._get_velocity base class method")
        return None

    def _put_velocity(self, start, vel):
        raise NotImplementedError("Fell through to TOD._put_velocity base class method")
        return

    # Read and write the common timestamps

    @function_timer
    def read_times(self, local_start=0, n=0, **kwargs):
        """Read timestamps.

        This reads the common set of timestamps that apply to all detectors
        in the TOD.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a numpy array containing the timestamps.

        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read times- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_times(local_start, n, **kwargs)

    @function_timer
    def write_times(self, local_start=0, stamps=None, **kwargs):
        """Write timestamps.

        This writes the common set of timestamps that apply to all detectors
        in the TOD.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            stamps (array): the array of timestamps to write.

        """
        if stamps is None:
            raise ValueError("you must specify the vector of time stamps")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write times- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + stamps.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + stamps.shape[0] - 1)
            )
        self._put_times(local_start, stamps, **kwargs)
        return

    # Read and write telescope boresight pointing

    @function_timer
    def read_boresight(self, local_start=0, n=0, **kwargs):
        """Read boresight quaternion pointing.

        This returns the pointing of the boresight in quaternions.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            A 2D array of shape (n, 4)

        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot read boresight- process has no local samples")
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_boresight(local_start, n, **kwargs)

    @function_timer
    def write_boresight(self, local_start=0, data=None, **kwargs):
        """Write boresight quaternion pointing.

        This writes the quaternion pointing for the boresight.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            data (array): 2D array of quaternions with shape[1] == 4.

        """
        if len(data.shape) != 2:
            raise ValueError("data should be a 2D array")
        if data.shape[1] != 4:
            raise ValueError("data should have second dimension of size 4")
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot write boresight- process has no local samples")
        if (local_start < 0) or (local_start + data.shape[0] > self.local_samples[1]):
            raise ValueError("local sample range is invalid")
        self._put_boresight(local_start, data, **kwargs)
        return

    @function_timer
    def read_boresight_azel(self, local_start=0, n=0, **kwargs):
        """Read boresight Azimuth / Elevation quaternion pointing.

        This returns the pointing of the boresight in the horizontal coordinate
        system, if it exists.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            A 2D array of shape (n, 4)

        Raises:
            NotImplementedError: if the telescope is not on the Earth.

        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot read boresight- process has no local samples")
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_boresight_azel(local_start, n, **kwargs)

    @function_timer
    def write_boresight_azel(self, local_start=0, data=None, **kwargs):
        """Write boresight Azimuth / Elevation quaternion pointing.

        This writes the quaternion pointing for the boresight in the horizontal
        coordinate system, if it exists.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            data (array): 2D array of quaternions with shape[1] == 4.

        Raises:
            RuntimeError or AttributeError : if the telescope is not on
                the Earth.

        """
        if len(data.shape) != 2:
            raise ValueError("data should be a 2D array")
        if data.shape[1] != 4:
            raise ValueError("data should have second dimension of size 4")
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot write boresight- process has no local samples")
        if (local_start < 0) or (local_start + data.shape[0] > self.local_samples[1]):
            raise ValueError("local sample range is invalid")
        self._put_boresight_azel(local_start, data, **kwargs)
        return

    # Read and write detector data

    @function_timer
    def read(self, detector=None, local_start=0, n=0, **kwargs):
        """Read detector data.

        This returns the timestream data for a single detector.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            An array containing the data.

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot read- process has no assigned local samples")
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get(detector, local_start, n, **kwargs)

    @function_timer
    def write(self, detector=None, local_start=0, data=None, **kwargs):
        """Write detector data.

        This writes the detector data.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            data (array): the data array.

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if data is None:
            raise ValueError("data array must be specified")
        if self.local_samples[1] <= 0:
            raise RuntimeError("cannot write- process has no assigned local samples")
        if (local_start < 0) or (local_start + data.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + data.shape[0] - 1)
            )
        self._put(detector, local_start, data, **kwargs)
        return

    # Read and write detector quaternion pointing

    @function_timer
    def read_pntg(self, detector=None, local_start=0, n=0, **kwargs):
        """Read detector quaternion pointing.

        This returns the pointing for a single detector in quaternions.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            A 2D array of shape (n, 4)

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read pntg- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_pntg(detector, local_start, n, **kwargs)

    @function_timer
    def write_pntg(self, detector=None, local_start=0, data=None, **kwargs):
        """Write detector quaternion pointing.

        This writes the quaternion pointing for a single detector.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            data (array): 2D array of quaternions with shape[1] == 4.

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if data is None:
            raise ValueError("data must be specified")
        if len(data.shape) != 2:
            raise ValueError("data should be a 2D array")
        if data.shape[1] != 4:
            raise ValueError("data should have second dimension of size 4")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write pntg- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + data.shape[0] > self.local_samples[1]):
            raise ValueError("local sample range is invalid")
        self._put_pntg(detector, local_start, data, **kwargs)
        return

    # Read and write detector flags

    @function_timer
    def read_flags(self, detector=None, local_start=0, n=0, **kwargs):
        """Read detector flags.

        This returns the detector-specific flags.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            An array containing the detector flags.

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read flags- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_flags(detector, local_start, n, **kwargs)

    @function_timer
    def read_common_flags(self, local_start=0, n=0, **kwargs):
        """Read common flags.

        This reads the common set of flags that should be applied to all
        detectors.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a numpy array containing the flags.

        """
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read common flags- process has no assigned local samples"
            )
        if n == 0:
            n = self.local_samples[1] - local_start
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_common_flags(local_start, n, **kwargs)

    @function_timer
    def write_common_flags(self, local_start=0, flags=None, **kwargs):
        """Write common flags.

        This writes the common set of flags that should be applied to all
        detectors.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            flags (array): array containing the flags to write.

        """
        if flags is None:
            raise ValueError("flags must be specified")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write common flags- process has no assigned local samples"
            )

        if (local_start < 0) or (local_start + flags.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + flags.shape[0] - 1)
            )
        self._put_common_flags(local_start, flags, **kwargs)
        return

    @function_timer
    def read_hwp_angle(self, local_start=0, n=0, **kwargs):
        """Read half-wave plate angle

        This reads the common HWP angle that should be applied to all
        detectors.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a numpy array containing the angles or None if the
                angle is not defined.

        """
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read HWP angle- process has no assigned local samples"
            )
        if n == 0:
            n = self.local_samples[1] - local_start
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        try:
            hwpangle = self._get_hwp_angle(local_start, n, **kwargs)
        except:
            hwpangle = None
        return hwpangle

    @function_timer
    def write_hwp_angle(self, local_start=0, hwpangle=None, **kwargs):
        """Write half-wave plate angle

        This writes the common HWP angle that should be applied to all
        detectors.

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            flags (array): array containing the flags to write.

        """
        if hwpangle is None:
            raise ValueError("hwpangle must be specified")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write HWP angle- process has no assigned local samples"
            )

        if (local_start < 0) or (local_start + flags.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + flags.shape[0] - 1)
            )
        self._put_hwp_angle(local_start, flags, **kwargs)
        return

    @function_timer
    def write_flags(self, detector=None, local_start=0, flags=None, **kwargs):
        """Write detector flags.

        This writes the detector-specific flags.

        Args:
            detector (str): the name of the detector.
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            flags (array): the detector flags.

        """
        if detector is None:
            raise ValueError("you must specify the detector")
        if detector not in self.local_dets:
            raise ValueError("detector {} not found".format(detector))
        if flags is None:
            raise ValueError("flags must be specified")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write flags- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + flags.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + flags.shape[0] - 1)
            )
        self._put_flags(detector, local_start, flags, **kwargs)
        return

    # Read and write telescope position

    @function_timer
    def read_position(self, local_start=0, n=0, **kwargs):
        """Read telescope position.

        This reads the telescope position in solar system barycenter
        coordinates (in Kilometers).

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a 2D numpy array containing the x,y,z coordinates at each
                sample.

        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read position- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_position(local_start, n, **kwargs)

    @function_timer
    def write_position(self, local_start=0, pos=None, **kwargs):
        """Write telescope position.

        This writes the telescope position in solar system barycenter
        coordinates (in Kilometers).

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            pos (array): the 2D array of x,y,z coordinates at each sample.

        """
        if pos is None:
            raise ValueError("you must specify the array of coordinates")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write position- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + pos.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + pos.shape[0] - 1)
            )
        self._put_position(local_start, pos, **kwargs)
        return

    # Read and write telescope velocity

    @function_timer
    def read_velocity(self, local_start=0, n=0, **kwargs):
        """Read telescope velocity.

        This reads the telescope velocity in solar system barycenter
        coordinates (in Kilometers/s).

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            n (int): the number of samples to read.  If zero, read to end.

        Returns:
            (array): a 2D numpy array containing the x,y,z velocity components
                at each sample.

        """
        if n == 0:
            n = self.local_samples[1] - local_start
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot read position- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + n > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + n - 1)
            )
        return self._get_velocity(local_start, n, **kwargs)

    @function_timer
    def write_velocity(self, local_start=0, vel=None, **kwargs):
        """Write telescope velocity.

        This writes the telescope velocity in solar system barycenter
        coordinates (in Kilometers/s).

        Args:
            local_start (int): the sample offset relative to the first locally
                assigned sample.
            vel (array): the 2D array of x,y,z velocity components at each
                sample.

        """
        if vel is None:
            raise ValueError("you must specify the array of velocities.")
        if self.local_samples[1] <= 0:
            raise RuntimeError(
                "cannot write times- process has no assigned local samples"
            )
        if (local_start < 0) or (local_start + vel.shape[0] > self.local_samples[1]):
            raise ValueError(
                "local sample range {} - {} is invalid"
                "".format(local_start, local_start + vel.shape[0] - 1)
            )
        self._put_velocity(local_start, vel, **kwargs)
        return


class TODCache(TOD):
    """TOD class that uses a memory cache for storage.

    This class simply uses a manually managed Cache object to store time
    ordered data.  You must "write" the data before you can "read" it.

    Args:
        mpicomm (mpi4py.MPI.Comm): the MPI communicator over which the
            data is distributed (or None).
        detectors (list):  The list of detector names.
        samples (int):  The total number of samples.
        detindx (dict): the detector indices for use in simulations.  Default is
            { x[0] : x[1] for x in zip(detectors, range(len(detectors))) }.
        detquats (dict):  Dictionary of detector quaternions.
        detranks (int):  The dimension of the process grid in the detector
            direction.  The MPI communicator size must be evenly divisible
            by this number.
        detbreaks (list):  Optional list of hard breaks in the detector
            distribution.
        sampsizes (list):  Optional list of sample chunk sizes which
            cannot be split.
        sampbreaks (list):  Optional list of hard breaks in the sample
            distribution.
    """

    def __init__(
        self,
        mpicomm,
        detectors,
        samples,
        detindx=None,
        detquats=None,
        detranks=1,
        detbreaks=None,
        sampsizes=None,
        sampbreaks=None,
    ):

        super().__init__(
            mpicomm,
            detectors,
            samples,
            detindx=detindx,
            detranks=detranks,
            detbreaks=detbreaks,
            sampsizes=sampsizes,
            sampbreaks=sampbreaks,
        )

        self._detquats = detquats
        self._pref_detdata = self.SIGNAL_NAME + "_"  # "toast_tod_detdata_"
        self._pref_detflags = self.FLAG_NAME + "_"  # "toast_tod_detflags_"
        self._pref_detpntg = "toast_tod_detpntg_"
        self._bore = "toast_boresight"
        self._bore_azel = "toast_boresight_azel"
        self._common = self.COMMON_FLAG_NAME  # "toast_tod_common_flags"
        self._stamps = self.TIMESTAMP_NAME  # "toast_tod_stamps"
        self._pos = "toast_tod_pos"
        self._vel = "toast_tod_vel"

    def detoffset(self):
        if self._detquats is None:
            raise NotImplementedError("TODCache does not contain detector quaternions.")
            return None
        else:
            return self._detquats

    # This class just uses a Cache object to store things.

    def _get(self, detector, start, n):
        if detector not in self.local_dets:
            raise ValueError(
                "detector {} not assigned to local process".format(detector)
            )
        cachedata = "{}{}".format(self._pref_detdata, detector)
        if not self.cache.exists(cachedata):
            raise ValueError("detector {} data not yet written".format(detector))
        dataref = self.cache.reference(cachedata)[start : start + n]
        return dataref

    def _put(self, detector, start, data):
        if detector not in self.local_dets:
            raise ValueError(
                "detector {} not assigned to local process".format(detector)
            )
        cachedata = "{}{}".format(self._pref_detdata, detector)

        if not self.cache.exists(cachedata):
            self.cache.create(cachedata, np.float64, (self.local_samples[1],))

        n = data.shape[0]
        refdata = self.cache.reference(cachedata)[start : start + n]

        refdata[:] = data
        return

    def _get_boresight(self, start, n):
        if not self.cache.exists(self._bore):
            raise ValueError("boresight not yet written")
        ref = self.cache.reference(self._bore)[start : start + n, :]
        return ref

    def _put_boresight(self, start, data):
        if not self.cache.exists(self._bore):
            self.cache.create(self._bore, np.float64, (self.local_samples[1], 4))
        ref = self.cache.reference(self._bore)
        ref[start : (start + data.shape[0]), :] = data
        return

    def _get_boresight_azel(self, start, n):
        if not self.cache.exists(self._bore_azel):
            raise ValueError("boresight not yet written")
        ref = self.cache.reference(self._bore_azel)[start : start + n, :]
        return ref

    def _put_boresight_azel(self, start, data):
        if not self.cache.exists(self._bore_azel):
            self.cache.create(self._bore_azel, np.float64, (self.local_samples[1], 4))
        ref = self.cache.reference(self._bore_azel)
        ref[start : (start + data.shape[0]), :] = data
        return

    def _get_pntg(self, detector, start, n):
        cachepntg = "{}{}".format(self._pref_detpntg, detector)
        if not self.cache.exists(cachepntg):
            # No detector-specific pointing written.  See if we have
            # boresight pointing and detector quaternions.
            if self.cache.exists(self._bore) and (self._detquats is not None):
                return qa.mult(
                    self.cache.reference(self._bore)[start : start + n, :],
                    self._detquats[detector],
                )
            else:
                raise ValueError(
                    "detector {}: pointing data not yet written, and boresight"
                    " and detector quaternions do not exist.".format(detector)
                )
        else:
            return self.cache.reference(cachepntg)[start : start + n, :]

    def _put_pntg(self, detector, start, data):
        if detector not in self.local_dets:
            raise ValueError(
                "detector {} not assigned to local process".format(detector)
            )
        cachepntg = "{}{}".format(self._pref_detpntg, detector)
        if not self.cache.exists(cachepntg):
            self.cache.create(cachepntg, np.float64, (self.local_samples[1], 4))
        pntgref = self.cache.reference(cachepntg)[start : (start + data.shape[0]), :]
        pntgref[:] = data
        return

    def _get_flags(self, detector, start, n):
        if detector not in self.local_dets:
            raise ValueError(
                "detector {} not assigned to local process".format(detector)
            )
        cacheflags = "{}{}".format(self._pref_detflags, detector)
        if not self.cache.exists(cacheflags):
            raise ValueError("detector {} flags not yet written".format(detector))
        flagsref = self.cache.reference(cacheflags)[start : start + n]
        return flagsref

    def _put_flags(self, detector, start, flags):
        if detector not in self.local_dets:
            raise ValueError(
                "detector {} not assigned to local process".format(detector)
            )
        cacheflags = "{}{}".format(self._pref_detflags, detector)

        if not self.cache.exists(cacheflags):
            self.cache.create(cacheflags, np.uint8, (self.local_samples[1],))

        n = flags.shape[0]
        refflags = self.cache.reference(cacheflags)[start : start + n]

        refflags[:] = flags
        return

    def _get_common_flags(self, start, n):
        if not self.cache.exists(self._common):
            raise ValueError("common flags not yet written")
        comref = self.cache.reference(self._common)[start : start + n]
        return comref

    def _put_common_flags(self, start, flags):
        if not self.cache.exists(self._common):
            self.cache.create(self._common, np.uint8, (self.local_samples[1],))
        n = flags.shape[0]
        comref = self.cache.reference(self._common)[start : start + n]
        comref[:] = flags
        return

    def _get_times(self, start, n):
        if not self.cache.exists(self._stamps):
            raise ValueError("timestamps not yet written")
        ref = self.cache.reference(self._stamps)[start : start + n]
        return ref

    def _put_times(self, start, stamps):
        if not self.cache.exists(self._stamps):
            self.cache.create(self._stamps, np.float64, (self.local_samples[1],))
        n = stamps.shape[0]
        ref = self.cache.reference(self._stamps)[start : start + n]
        ref[:] = stamps
        return

    def _get_position(self, start, n):
        if not self.cache.exists(self._pos):
            raise ValueError("telescope position not yet written")
        ref = self.cache.reference(self._pos)[start : start + n]
        return ref

    def _put_position(self, start, pos):
        if not self.cache.exists(self._pos):
            self.cache.create(self._pos, np.float64, (self.local_samples[1], 3))
        n = pos.shape[0]
        ref = self.cache.reference(self._pos)[start : start + n, :]
        ref[:, :] = pos
        return

    def _get_velocity(self, start, n):
        if not self.cache.exists(self._vel):
            raise ValueError("telescope velocity not yet written")
        ref = self.cache.reference(self._vel)[start : start + n]
        return ref

    def _put_velocity(self, start, vel):
        if not self.cache.exists(self._vel):
            self.cache.create(self._vel, np.float64, (self.local_samples[1], 3))
        n = vel.shape[0]
        ref = self.cache.reference(self._vel)[start : start + n, :]
        ref[:, :] = vel
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.


available = True
try:
    import tidas as tds
    from tidas.mpi import MPIVolume
except ImportError:
    available = False


def to_dict(td):
    meta = dict()
    for k in td.keys():
        tp = td.get_type(k)
        if tp == "d":
            meta[k] = td.get_float64(k)
        elif tp == "f":
            meta[k] = td.get_float32(k)
        elif tp == "l":
            meta[k] = td.get_int64(k)
        elif tp == "L":
            meta[k] = td.get_uint64(k)
        elif tp == "i":
            meta[k] = td.get_int32(k)
        elif tp == "I":
            meta[k] = td.get_uint32(k)
        elif tp == "h":
            meta[k] = td.get_int16(k)
        elif tp == "H":
            meta[k] = td.get_uint16(k)
        elif tp == "b":
            meta[k] = td.get_int8(k)
        elif tp == "B":
            meta[k] = td.get_uint8(k)
        else:
            meta[k] = td.get_string(k)
    return meta


def from_dict(meta):
    td = tds.Dictionary()
    for k, v in meta.items():
        if isinstance(v, float):
            td.put_float64(k, v)
        elif isinstance(v, int):
            td.put_int64(k, v)
        else:
            td.put_string(k, str(v))
    return td


def find_obs(vol, parent, name):
    """Return the TIDAS block representing an observation.

    If the block exists, a handle is returned.  Otherwise it is created.

    Args:
        vol (tidas.MPIVolume):  the volume.
        parent (str):  the path to the parent block.
        name (str):  the name of the observation block.

    Returns:
        (tidas.Block):  the block handle

    """
    # The root block
    root = vol.root()

    par = root
    if parent != "":
        # Descend tree to the parent node
        parentnodes = parent.split("/")
        for pn in parentnodes:
            if pn != "":
                par = par.block_get(pn)
    obs = None
    if name in par.block_names():
        obs = par.block_get(name)
    else:
        # Create the observation block
        obs = par.block_add(name, tds.Block())
    del par
    del root
    return obs



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ..timing import function_timer

from .interval import Interval


@function_timer
def regular_intervals(n, start, first, rate, duration, gap):
    """Function to generate simulated regular intervals.

    This creates a list of intervals, given a start time/sample and time
    span for the interval and the gap in time between intervals.  The
    length of the interval and the total interval + gap are rounded down to the
    nearest sample and all intervals in the list are created using those
    lengths.

    If the time span is an exact multiple of the sampling, then the
    final sample is excluded.  The reason we always round down to the whole
    number of samples that fits inside the time range is so that the requested
    time span boundary (one hour, one day, etc) will fall in between the last
    sample of one interval and the first sample of the next.

    Example:  you want to simulate science observations of length 22 hours and
    then have 4 hours of down time (e.g. a cooler cycle).  Specifying a
    duration of 22*3600 and a gap of 4*3600 will result in a total time for the
    science + gap of a fraction of a sample less than 26 hours.  So the
    requested 26 hour mark will fall between the last sample of one regular
    interval and the first sample of the next.  Note that this fraction of
    a sample will accumulate if you have many, many intervals.

    This function is intended only for simulations- in the case of real data,
    the timestamp of every sample is known and boundaries between changes in
    the experimental configuration are already specified.

    Args:
        n (int): the number of intervals.
        start (float): the start time in seconds.
        first (int): the first sample index, which occurs at "start".
        rate (float): the sample rate in Hz.
        duration (float): the length of the interval in seconds.
        gap (float): the length of the gap in seconds.

    Returns:
        (list): a list of Interval objects.

    """
    invrate = 1.0 / rate

    # Compute the whole number of samples that fit within the
    # requested time span (rounded down to a whole number).  Check for the
    # case of the time span being an exact number of samples- in which case
    # the final sample is excluded.

    lower = int((duration + gap) * rate)
    totsamples = None
    if np.absolute(lower * invrate - (duration + gap)) > 1.0e-12:
        totsamples = lower + 1
    else:
        totsamples = lower

    lower = int(duration * rate)
    dursamples = None
    if np.absolute(lower * invrate - duration) > 1.0e-12:
        dursamples = lower + 1
    else:
        dursamples = lower

    gapsamples = totsamples - dursamples

    intervals = []

    for i in range(n):
        ifirst = first + i * totsamples
        ilast = ifirst + dursamples - 1
        # The time span between interval starts (the first sample of one
        # interval to the first sample of the next) includes the one extra
        # sample time.
        istart = start + i * (totsamples * invrate)
        # The stop time is the timestamp of the last valid sample (thus the -1).
        istop = istart + ((dursamples - 1) * invrate)
        intervals.append(Interval(start=istart, stop=istop, first=ifirst, last=ilast))

    return intervals



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

from ..mpi import MPI, use_mpi

import os

import healpy as hp
import numpy as np

from ..cache import Cache
from ..op import Operator
from ..timing import function_timer, Timer
from ..utils import Logger, memreport

madam = None
if use_mpi:
    try:
        import libmadam_wrapper as madam
    except ImportError:
        madam = None


def count_caches(data, comm, nodecomm, madamcache, msg=""):
    """Count the amount of memory in the TOD caches"""
    my_todsize = 0
    for obs in data.obs:
        tod = obs["tod"]
        my_todsize += tod.cache.report(silent=True)
    my_cachesize = madamcache.report(silent=True)
    node_todsize = nodecomm.allreduce(my_todsize, MPI.SUM)
    node_cachesize = nodecomm.allreduce(my_cachesize, MPI.SUM)
    if comm.rank == 0:
        print(
            "Node has {:.3f} GB allocated in TOAST TOD caches and "
            "{:.3f} GB in Madam caches ({:.3f} GB total) {}".format(
                node_todsize / 2 ** 30,
                node_cachesize / 2 ** 30,
                (node_todsize + node_cachesize) / 2 ** 30,
                msg,
            )
        )
    return


class OpMadam(Operator):
    """Operator which passes data to libmadam for map-making.

    Args:
        params (dictionary): parameters to pass to madam.
        detweights (dictionary): individual noise weights to use for each
            detector.
        pixels (str): the name of the cache object (<pixels>_<detector>)
            containing the pixel indices to use.
        pixels_nested (bool): Set to False if the pixel numbers are in
            ring ordering. Default is True.
        weights (str): the name of the cache object (<weights>_<detector>)
            containing the pointing weights to use.
        name (str): the name of the cache object (<name>_<detector>) to
            use for the detector timestream.  If None, use the TOD.
        name_out (str): the name of the cache object (<name>_<detector>)
            to use to output destriped detector timestream.
            No output if None.
        flag_name (str): the name of the cache object
            (<flag_name>_<detector>) to use for the detector flags.
            If None, use the TOD.
        flag_mask (int): the integer bit mask (0-255) that should be
            used with the detector flags in a bitwise AND.
        common_flag_name (str): the name of the cache object
            to use for the common flags.  If None, use the TOD.
        common_flag_mask (int): the integer bit mask (0-255) that should
            be used with the common flags in a bitwise AND.
        apply_flags (bool): whether to apply flags to the pixel numbers.
        purge (bool): if True, clear any cached data that is copied into
            the Madam buffers.
        purge_tod (bool): if True, clear any cached signal that is
            copied into the Madam buffers.
        purge_pixels (bool): if True, clear any cached pixels that are
            copied into the Madam buffers.
        purge_weights (bool): if True, clear any cached weights that are
            copied into the Madam buffers.
        purge_flags (bool): if True, clear any cached flags that are
            copied into the Madam buffers.
        dets (iterable):  List of detectors to map. If left as None, all
            available detectors are mapped.
        mcmode (bool): If true, the operator is constructed in
            Monte Carlo mode and Madam will cache auxiliary information
            such as pixel matrices and noise filter.
        noise (str): Keyword to use when retrieving the noise object
            from the observation.
        conserve_memory(bool/int): Stagger the Madam buffer staging on node.
        translate_timestamps(bool): Translate timestamps to enforce
            monotonity.

    """

    def __init__(
        self,
        params={},
        detweights=None,
        pixels="pixels",
        pixels_nested=True,
        weights="weights",
        name=None,
        name_out=None,
        flag_name=None,
        flag_mask=255,
        common_flag_name=None,
        common_flag_mask=255,
        apply_flags=True,
        purge=False,
        dets=None,
        mcmode=False,
        purge_tod=False,
        purge_pixels=False,
        purge_weights=False,
        purge_flags=False,
        noise="noise",
        intervals="intervals",
        conserve_memory=True,
        translate_timestamps=True,
    ):
        # Call the parent class constructor
        super().__init__()

        # madam uses time-based distribution
        self._name = name
        self._name_out = name_out
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._pixels = pixels
        self._pixels_nested = pixels_nested
        self._weights = weights
        self._detw = detweights
        self._purge = purge
        if self._purge:
            self._purge_tod = True
            self._purge_pixels = True
            self._purge_weights = True
            self._purge_flags = True
        else:
            self._purge_tod = purge_tod
            self._purge_pixels = purge_pixels
            self._purge_weights = purge_weights
            self._purge_flags = purge_flags
        self._apply_flags = apply_flags
        self.params = params
        if dets is not None:
            self._dets = set(dets)
        else:
            self._dets = None
        self._mcmode = mcmode
        if mcmode:
            self.params["mcmode"] = True
        else:
            self.params["mcmode"] = False
        if self._name_out is not None:
            self.params["write_tod"] = True
        else:
            self.params["write_tod"] = False
        self._cached = False
        self._noisekey = noise
        self._intervals = intervals
        self._cache = Cache()
        self._madam_timestamps = None
        self._madam_pixels = None
        self._madam_pixweights = None
        self._madam_signal = None
        if conserve_memory is None:
            conserve_memory = True
        self._conserve_memory = int(conserve_memory)
        self._translate_timestamps = translate_timestamps
        if "info" in params:
            self._verbose = int(params["info"]) > 0
        else:
            self._verbose = True

    def __del__(self):
        self._cache.clear()
        if self._cached:
            madam.clear_caches()
            self._cached = False

    @property
    def available(self):
        """(bool): True if libmadam is found in the library search path."""
        return madam is not None and madam.available

    @function_timer
    def exec(self, data, comm=None):
        """Copy data to Madam-compatible buffers and make a map.

        Args:
            data (toast.Data): The distributed data.

        Returns:
            None

        """
        if not self.available:
            raise RuntimeError("libmadam is not available")

        if len(data.obs) == 0:
            raise RuntimeError(
                "OpMadam requires every supplied data object to "
                "contain at least one observation"
            )

        # User is able to adjust verbosity between calls to OpMadam.exec()
        if "info" in self.params:
            self._verbose = int(self.params["info"]) > 0

        if comm is None:
            # Use the world communicator from the distributed data.
            comm = data.comm.comm_world
        self._data = data
        self._comm = comm
        self._rank = comm.rank

        (
            pars,
            dets,
            nsamp,
            ndet,
            nnz,
            nnz_full,
            nnz_stride,
            periods,
            obs_period_ranges,
            psdfreqs,
            nside,
        ) = self._prepare()

        psdinfo, signal_type, pixels_dtype, weight_dtype = self._stage_data(
            nsamp,
            ndet,
            nnz,
            nnz_full,
            nnz_stride,
            obs_period_ranges,
            psdfreqs,
            dets,
            nside,
        )

        self._destripe(pars, dets, periods, psdinfo)

        self._unstage_data(
            nsamp,
            nnz,
            nnz_full,
            obs_period_ranges,
            dets,
            signal_type,
            pixels_dtype,
            nside,
            weight_dtype,
        )

        return

    @function_timer
    def _destripe(self, pars, dets, periods, psdinfo):
        """Destripe the buffered data"""
        if self._verbose:
            memreport("just before calling libmadam.destripe", self._comm)
        if self._cached:
            # destripe
            outpath = ""
            if "path_output" in self.params:
                outpath = self.params["path_output"]
            outpath = outpath.encode("ascii")
            madam.destripe_with_cache(
                self._comm,
                self._madam_timestamps,
                self._madam_pixels,
                self._madam_pixweights,
                self._madam_signal,
                outpath,
            )
        else:
            (detweights, npsd, psdstarts, psdfreqs, psdvals) = psdinfo

            # destripe
            madam.destripe(
                self._comm,
                pars,
                dets,
                detweights,
                self._madam_timestamps,
                self._madam_pixels,
                self._madam_pixweights,
                self._madam_signal,
                periods,
                npsd,
                psdstarts,
                psdfreqs,
                psdvals,
            )

            if self._mcmode:
                self._cached = True
        return

    def _count_samples(self):
        """Loop over the observations and count the number of samples."""
        if len(self._data.obs) != 1:
            nsamp = 0
            tod0 = self._data.obs[0]["tod"]
            detectors0 = tod0.local_dets
            for obs in self._data.obs:
                tod = obs["tod"]
                # For the moment, we require that all observations have
                # the same set of detectors
                detectors = tod.local_dets
                dets_are_same = True
                if len(detectors0) != len(detectors):
                    dets_are_same = False
                else:
                    for det1, det2 in zip(detectors0, detectors):
                        if det1 != det2:
                            dets_are_same = False
                            break
                if not dets_are_same:
                    raise RuntimeError(
                        "When calling Madam, all TOD assigned to a process "
                        "must have the same local detectors."
                    )
                nsamp += tod.local_samples[1]
        else:
            tod = self._data.obs[0]["tod"]
            nsamp = tod.local_samples[1]
        return nsamp

    def _get_period_ranges(self, detectors, nsamp):
        """Collect the ranges of every observation."""
        log = Logger.get()
        timer = Timer()
        timer.start()
        # Discard intervals that are too short to fit a baseline
        if "basis_order" in self.params:
            norder = int(self.params["basis_order"]) + 1
        else:
            norder = 1

        psdfreqs = None
        period_lengths = []
        obs_period_ranges = []

        for obs in self._data.obs:
            tod = obs["tod"]
            # Check that all noise objects have the same binning
            if self._noisekey in obs.keys():
                nse = obs[self._noisekey]
                if nse is not None:
                    if psdfreqs is None:
                        psdfreqs = nse.freq(detectors[0]).astype(np.float64).copy()
                    for det in detectors:
                        check_psdfreqs = nse.freq(det)
                        if not np.allclose(psdfreqs, check_psdfreqs):
                            raise RuntimeError(
                                "All PSDs passed to Madam must have"
                                " the same frequency binning."
                            )
            # Collect the valid intervals for this observation
            period_ranges = []
            if self._intervals in obs:
                intervals = obs[self._intervals]
            else:
                intervals = None
            local_intervals = tod.local_intervals(intervals)

            for ival in local_intervals:
                local_start = ival.first
                local_stop = ival.last + 1
                if local_stop - local_start < norder:
                    continue
                period_lengths.append(local_stop - local_start)
                period_ranges.append((local_start, local_stop))
            obs_period_ranges.append(period_ranges)

        # Update the number of samples based on the valid intervals

        nsamp_tot_full = self._comm.allreduce(nsamp, op=MPI.SUM)
        nperiod = len(period_lengths)
        period_lengths = np.array(period_lengths, dtype=np.int64)
        nsamp = np.sum(period_lengths, dtype=np.int64)
        nsamp_tot = self._comm.allreduce(nsamp, op=MPI.SUM)
        if nsamp_tot == 0:
            raise RuntimeError(
                "No samples in valid intervals: nsamp_tot_full = {}, "
                "nsamp_tot = {}".format(nsamp_tot_full, nsamp_tot)
            )
        if self._verbose and self._rank == 0:
            log.info(
                "OpMadam: {:.2f} % of samples are included in valid "
                "intervals.".format(nsamp_tot * 100.0 / nsamp_tot_full)
            )

        # Madam expects starting indices, not period lengths
        periods = np.zeros(nperiod, dtype=np.int64)
        for i, n in enumerate(period_lengths[:-1]):
            periods[i + 1] = periods[i] + n

        self._comm.Barrier()
        if self._verbose and self._rank == 0:
            timer.report_clear("Collect period ranges")

        return obs_period_ranges, psdfreqs, periods, nsamp

    @function_timer
    def _prepare(self):
        """Examine the data object."""
        log = Logger.get()
        timer = Timer()
        timer.start()

        nsamp = self._count_samples()

        # Determine the detectors and the pointing matrix non-zeros
        # from the first observation. Madam will expect these to remain
        # unchanged across observations.

        tod = self._data.obs[0]["tod"]

        if self._dets is None:
            dets = tod.local_dets
        else:
            dets = [det for det in tod.local_dets if det in self._dets]
        ndet = len(dets)

        # to get the number of Non-zero pointing weights per pixel,
        # we use the fact that for Madam, all processes have all detectors
        # for some slice of time.  So we can get this information from the
        # shape of the data from the first detector

        nnzname = "{}_{}".format(self._weights, dets[0])
        nnz_full = tod.cache.reference(nnzname).shape[1]

        if "temperature_only" in self.params and self.params["temperature_only"] in [
            "T",
            "True",
            "TRUE",
            "true",
            True,
        ]:
            if nnz_full not in [1, 3]:
                raise RuntimeError(
                    "OpMadam: Don't know how to make a temperature map "
                    "with nnz={}".format(nnz_full)
                )
            nnz = 1
            nnz_stride = nnz_full
        else:
            nnz = nnz_full
            nnz_stride = 1

        if "nside_map" not in self.params:
            raise RuntimeError(
                'OpMadam: "nside_map" must be set in the parameter dictionary'
            )
        nside = int(self.params["nside_map"])

        if self._rank == 0 and "path_output" in self.params:
            os.makedirs(self.params["path_output"], exist_ok=True)

        # Inspect the valid intervals across all observations to
        # determine the number of samples per detector

        obs_period_ranges, psdfreqs, periods, nsamp = self._get_period_ranges(
            dets, nsamp
        )

        self._comm.Barrier()
        if self._rank == 0 and self._verbose:
            timer.report_clear("Collect dataset dimensions")

        return (
            self.params,
            dets,
            nsamp,
            ndet,
            nnz,
            nnz_full,
            nnz_stride,
            periods,
            obs_period_ranges,
            psdfreqs,
            nside,
        )

    @function_timer
    def _stage_time(self, detectors, nsamp, obs_period_ranges):
        """Stage the timestamps and use them to build PSD inputs."""
        self._madam_timestamps = self._cache.create(
            "timestamps", madam.TIMESTAMP_TYPE, (nsamp,)
        )

        offset = 0
        time_offset = 0
        psds = {}
        for iobs, obs in enumerate(self._data.obs):
            tod = obs["tod"]
            period_ranges = obs_period_ranges[iobs]

            # Collect the timestamps for the valid intervals
            timestamps = tod.local_times().copy()
            if self._translate_timestamps:
                # Translate the time stamps to be monotonous
                timestamps -= timestamps[0] - time_offset
                time_offset = timestamps[-1] + 1

            for istart, istop in period_ranges:
                nn = istop - istart
                ind = slice(offset, offset + nn)
                self._madam_timestamps[ind] = timestamps[istart:istop]
                offset += nn

            # get the noise object for this observation and create new
            # entries in the dictionary when the PSD actually changes
            if self._noisekey in obs.keys():
                nse = obs[self._noisekey]
                if "noise_scale" in obs:
                    noise_scale = obs["noise_scale"]
                else:
                    noise_scale = 1
                if nse is not None:
                    for det in detectors:
                        psd = nse.psd(det) * noise_scale ** 2
                        if det not in psds:
                            psds[det] = [(0, psd)]
                        else:
                            if not np.allclose(psds[det][-1][1], psd):
                                psds[det] += [(timestamps[0], psd)]

        return psds

    @function_timer
    def _stage_signal(self, detectors, nsamp, ndet, obs_period_ranges, nodecomm, nread):
        """Stage signal"""
        log = Logger.get()
        timer = Timer()
        # Determine if we can purge the signal and avoid keeping two
        # copies in memory
        purge = self._name is not None and (
            self._purge_tod or self._name == self._name_out
        )
        if not purge:
            nread = 1
            nodecomm = MPI.COMM_SELF

        for iread in range(nread):
            nodecomm.Barrier()
            timer.start()
            if nodecomm.rank % nread == iread:
                # Allocate Madam buffer
                self._madam_signal = self._cache.create(
                    "signal", madam.SIGNAL_TYPE, (nsamp * ndet,)
                )
                self._madam_signal[:] = np.nan
                # Fill Madam buffer
                global_offset = 0
                for iobs, obs in enumerate(self._data.obs):
                    tod = obs["tod"]
                    period_ranges = obs_period_ranges[iobs]
                    for idet, det in enumerate(detectors):
                        # Get the signal.
                        signal = tod.local_signal(det, self._name)
                        signal_dtype = signal.dtype
                        offset = global_offset
                        for istart, istop in period_ranges:
                            nn = istop - istart
                            dslice = slice(
                                idet * nsamp + offset, idet * nsamp + offset + nn
                            )
                            self._madam_signal[dslice] = signal[istart:istop]
                            offset += nn
                        del signal
                    # Purge only after all detectors are staged in case some are aliased
                    # cache.clear() will not fail if the object was already
                    # deleted as an alias
                    if purge:
                        for det in detectors:
                            cachename = "{}_{}".format(self._name, det)
                            tod.cache.clear(cachename)
                    global_offset = offset
            if self._verbose and nread > 1:
                nodecomm.Barrier()
                if self._rank == 0:
                    timer.report_clear("Stage signal {} / {}".format(iread + 1, nread))
        return signal_dtype

    @function_timer
    def _stage_pixels(self, detectors, nsamp, ndet, obs_period_ranges, nside):
        """Stage pixels"""
        self._madam_pixels = self._cache.create(
            "pixels", madam.PIXEL_TYPE, (nsamp * ndet,)
        )
        self._madam_pixels[:] = -1

        global_offset = 0
        for iobs, obs in enumerate(self._data.obs):
            tod = obs["tod"]
            period_ranges = obs_period_ranges[iobs]

            commonflags = None
            for idet, det in enumerate(detectors):
                # Optionally get the flags, otherwise they are
                # assumed to have been applied to the pixel numbers.

                if self._apply_flags:
                    detflags = tod.local_flags(det, self._flag_name)
                    commonflags = tod.local_common_flags(self._common_flag_name)
                    flags = np.logical_or(
                        (detflags & self._flag_mask) != 0,
                        (commonflags & self._common_flag_mask) != 0,
                    )
                    del detflags

                # get the pixels for the valid intervals from the cache

                pixelsname = "{}_{}".format(self._pixels, det)
                pixels = tod.cache.reference(pixelsname)
                pixels_dtype = pixels.dtype

                if not self._pixels_nested:
                    # Madam expects the pixels to be in nested ordering
                    pixels = pixels.copy()
                    good = pixels >= 0
                    pixels[good] = hp.ring2nest(nside, pixels[good])

                if self._apply_flags:
                    pixels = pixels.copy()
                    pixels[flags] = -1

                offset = global_offset
                for istart, istop in period_ranges:
                    nn = istop - istart
                    dslice = slice(idet * nsamp + offset, idet * nsamp + offset + nn)
                    self._madam_pixels[dslice] = pixels[istart:istop]
                    offset += nn

                del pixels
                if self._apply_flags:
                    del flags

            # Always purge the pixels but restore them from the Madam
            # buffers when purge_pixels=False
            # Purging MUST happen after all detectors are staged because
            # some the pixel numbers may be aliased between detectors
            for det in detectors:
                pixelsname = "{}_{}".format(self._pixels, det)
                # cache.clear() will not fail if the object was already
                # deleted as an alias
                tod.cache.clear(pixelsname)
                if self._purge_flags and self._flag_name is not None:
                    cacheflagname = "{}_{}".format(self._flag_name, det)
                    tod.cache.clear(cacheflagname)

            del commonflags
            if self._purge_flags and self._common_flag_name is not None:
                tod.cache.clear(self._common_flag_name)
            global_offset = offset
        return pixels_dtype

    @function_timer
    def _stage_pixweights(
        self,
        detectors,
        nsamp,
        ndet,
        nnz,
        nnz_full,
        nnz_stride,
        obs_period_ranges,
        nodecomm,
        nread,
    ):
        """Now collect the pixel weights"""
        log = Logger.get()
        timer = Timer()
        # Determine if we can purge the pixel weights and avoid keeping two
        # copies of the weights in memory
        purge = self._purge_weights or (nnz == nnz_full)
        if not purge:
            nread = 1
            nodecomm = MPI.COMM_SELF

        for iread in range(nread):
            nodecomm.Barrier()
            timer.start()
            if nodecomm.rank % nread == iread:
                # Allocate Madam buffer
                self._madam_pixweights = self._cache.create(
                    "pixweights", madam.WEIGHT_TYPE, (nsamp * ndet * nnz,)
                )
                self._madam_pixweights[:] = 0
                # Fill Madam buffer
                global_offset = 0
                for iobs, obs in enumerate(self._data.obs):
                    tod = obs["tod"]
                    period_ranges = obs_period_ranges[iobs]
                    for idet, det in enumerate(detectors):
                        # get the pixels and weights for the valid intervals
                        # from the cache
                        weightsname = "{}_{}".format(self._weights, det)
                        weights = tod.cache.reference(weightsname)
                        weight_dtype = weights.dtype
                        offset = global_offset
                        for istart, istop in period_ranges:
                            nn = istop - istart
                            dwslice = slice(
                                (idet * nsamp + offset) * nnz,
                                (idet * nsamp + offset + nn) * nnz,
                            )
                            self._madam_pixweights[dwslice] = weights[
                                istart:istop
                            ].flatten()[::nnz_stride]
                            offset += nn
                        del weights
                    # Purge the weights but restore them from the Madam
                    # buffers when purge_weights=False.
                    if purge:
                        for idet, det in enumerate(detectors):
                            weightsname = "{}_{}".format(self._weights, det)
                            tod.cache.clear(weightsname)
                    global_offset = offset
            if self._verbose and nread > 1:
                nodecomm.Barrier()
                if self._rank == 0:
                    timer.report_clear(
                        "Stage pixel weights {} / {}".format(iread + 1, nread)
                    )
        return weight_dtype

    @function_timer
    def _stage_data(
        self,
        nsamp,
        ndet,
        nnz,
        nnz_full,
        nnz_stride,
        obs_period_ranges,
        psdfreqs,
        detectors,
        nside,
    ):
        """create madam-compatible buffers

        Collect the TOD into Madam buffers. Process pixel weights
        Separate from the rest to reduce the memory high water mark
        When the user has set purge=True

        Moving data between toast and Madam buffers has an overhead.
        We perform the operation in a staggered fashion to have the
        overhead only once per node.

        """
        log = Logger.get()
        nodecomm = self._comm.Split_type(MPI.COMM_TYPE_SHARED, self._rank)
        # Check if the user has elected to stagger staging the data on each
        # node to avoid exhausting memory
        if self._conserve_memory:
            if self._conserve_memory == 1:
                nread = nodecomm.size
            else:
                nread = min(self._conserve_memory, nodecomm.size)
        else:
            nread = 1

        self._comm.Barrier()
        timer_tot = Timer()
        timer_tot.start()

        # Stage time, it is never purged so the staging is never stepped
        timer = Timer()
        timer.start()
        psds = self._stage_time(detectors, nsamp, obs_period_ranges)
        if self._verbose:
            nodecomm.Barrier()
            if self._rank == 0:
                timer.report_clear("Stage time")
        memreport("after staging time", self._comm)  # DEBUG
        count_caches(
            self._data, self._comm, nodecomm, self._cache, "after staging time"
        )  # DEBUG

        # Stage signal.  If signal is not being purged, staging is not stepped
        timer.start()
        signal_dtype = self._stage_signal(
            detectors, nsamp, ndet, obs_period_ranges, nodecomm, nread
        )
        if self._verbose:
            nodecomm.Barrier()
            if self._rank == 0:
                timer.report_clear("Stage signal")
        memreport("after staging signal", self._comm)  # DEBUG
        count_caches(
            self._data, self._comm, nodecomm, self._cache, "after staging signal"
        )  # DEBUG

        # Stage pixels
        timer_step = Timer()
        timer_step.start()
        for iread in range(nread):
            nodecomm.Barrier()
            timer.start()
            if nodecomm.rank % nread == iread:
                pixels_dtype = self._stage_pixels(
                    detectors, nsamp, ndet, obs_period_ranges, nside
                )
            if self._verbose and nread > 1:
                nodecomm.Barrier()
                if self._rank == 0:
                    timer.report_clear("Stage pixels {} / {}".format(iread + 1, nread))
        if self._verbose:
            nodecomm.Barrier()
            if self._rank == 0:
                timer_step.report_clear("Stage pixels")
        memreport("after staging pixels", self._comm)  # DEBUG
        count_caches(
            self._data, self._comm, nodecomm, self._cache, "after staging pixels"
        )  # DEBUG

        # Stage pixel weights
        timer_step.start()
        weight_dtype = self._stage_pixweights(
            detectors,
            nsamp,
            ndet,
            nnz,
            nnz_full,
            nnz_stride,
            obs_period_ranges,
            nodecomm,
            nread,
        )
        if self._verbose:
            nodecomm.Barrier()
            if self._rank == 0:
                timer_step.report_clear("Stage pixel weights")
        memreport("after staging pixel weights", self._comm)  # DEBUG
        count_caches(
            self._data, self._comm, nodecomm, self._cache, "after staging pixel weights"
        )  # DEBUG

        del nodecomm
        if self._rank == 0 and self._verbose:
            timer_tot.report_clear("Stage all data")

        # detweights is either a dictionary of weights specified at
        # construction time, or else we use uniform weighting.
        detw = {}
        if self._detw is None:
            for idet, det in enumerate(detectors):
                detw[det] = 1.0
        else:
            detw = self._detw

        detweights = np.zeros(ndet, dtype=np.float64)
        for idet, det in enumerate(detectors):
            detweights[idet] = detw[det]

        if len(psds) > 0:
            npsdbin = len(psdfreqs)

            npsd = np.zeros(ndet, dtype=np.int64)
            psdstarts = []
            psdvals = []
            for idet, det in enumerate(detectors):
                if det not in psds:
                    raise RuntimeError("Every detector must have at least " "one PSD")
                psdlist = psds[det]
                npsd[idet] = len(psdlist)
                for psdstart, psd in psdlist:
                    psdstarts.append(psdstart)
                    psdvals.append(psd)
            npsdtot = np.sum(npsd)
            psdstarts = np.array(psdstarts, dtype=np.float64)
            psdvals = np.hstack(psdvals).astype(madam.PSD_TYPE)
            npsdval = psdvals.size
        else:
            npsd = np.ones(ndet, dtype=np.int64)
            npsdtot = np.sum(npsd)
            psdstarts = np.zeros(npsdtot)
            npsdbin = 10
            fsample = 10.0
            psdfreqs = np.arange(npsdbin) * fsample / npsdbin
            npsdval = npsdbin * npsdtot
            psdvals = np.ones(npsdval)
        psdinfo = (detweights, npsd, psdstarts, psdfreqs, psdvals)
        if self._rank == 0 and self._verbose:
            timer_tot.report_clear("Collect PSD info")
        return psdinfo, signal_dtype, pixels_dtype, weight_dtype

    @function_timer
    def _unstage_signal(self, detectors, nsamp, obs_period_ranges, signal_type):
        if self._name_out is not None:
            global_offset = 0
            for obs, period_ranges in zip(self._data.obs, obs_period_ranges):
                tod = obs["tod"]
                nlocal = tod.local_samples[1]
                for idet, det in enumerate(detectors):
                    signal = np.ones(nlocal, dtype=signal_type) * np.nan
                    offset = global_offset
                    for istart, istop in period_ranges:
                        nn = istop - istart
                        dslice = slice(
                            idet * nsamp + offset, idet * nsamp + offset + nn
                        )
                        signal[istart:istop] = self._madam_signal[dslice]
                        offset += nn
                    cachename = "{}_{}".format(self._name_out, det)
                    tod.cache.put(cachename, signal, replace=True)
                global_offset = offset
        self._madam_signal = None
        self._cache.destroy("signal")
        return

    @function_timer
    def _unstage_pixels(self, detectors, nsamp, obs_period_ranges, pixels_dtype, nside):
        if not self._purge_pixels:
            # restore the pixels from the Madam buffers
            global_offset = 0
            for obs, period_ranges in zip(self._data.obs, obs_period_ranges):
                tod = obs["tod"]
                nlocal = tod.local_samples[1]
                for idet, det in enumerate(detectors):
                    pixels = -(np.ones(nlocal, dtype=pixels_dtype))
                    offset = global_offset
                    for istart, istop in period_ranges:
                        nn = istop - istart
                        dslice = slice(
                            idet * nsamp + offset, idet * nsamp + offset + nn
                        )
                        pixels[istart:istop] = self._madam_pixels[dslice]
                        offset += nn
                    npix = 12 * nside ** 2
                    good = np.logical_and(pixels >= 0, pixels < npix)
                    if not self._pixels_nested:
                        pixels[good] = hp.nest2ring(nside, pixels[good])
                    pixels[np.logical_not(good)] = -1
                    cachename = "{}_{}".format(self._pixels, det)
                    tod.cache.put(cachename, pixels, replace=True)
                global_offset = offset
        self._madam_pixels = None
        self._cache.destroy("pixels")
        return

    @function_timer
    def _unstage_pixweights(
        self, detectors, nsamp, obs_period_ranges, weight_dtype, nnz, nnz_full
    ):
        if not self._purge_weights and nnz == nnz_full:
            # restore the weights from the Madam buffers
            global_offset = 0
            for obs, period_ranges in zip(self._data.obs, obs_period_ranges):
                tod = obs["tod"]
                nlocal = tod.local_samples[1]
                for idet, det in enumerate(detectors):
                    weights = np.zeros([nlocal, nnz], dtype=weight_dtype)
                    offset = global_offset
                    for istart, istop in period_ranges:
                        nn = istop - istart
                        dwslice = slice(
                            (idet * nsamp + offset) * nnz,
                            (idet * nsamp + offset + nn) * nnz,
                        )
                        weights[istart:istop] = self._madam_pixweights[dwslice].reshape(
                            [-1, nnz]
                        )
                        offset += nn
                    cachename = "{}_{}".format(self._weights, det)
                    tod.cache.put(cachename, weights, replace=True)
                global_offset = offset
        self._madam_pixweights = None
        self._cache.destroy("pixweights")
        return

    @function_timer
    def _unstage_data(
        self,
        nsamp,
        nnz,
        nnz_full,
        obs_period_ranges,
        detectors,
        signal_type,
        pixels_dtype,
        nside,
        weight_dtype,
    ):
        """Clear Madam buffers, restore pointing into TOAST caches
        and cache the destriped signal.

        """
        log = Logger.get()
        self._madam_timestamps = None
        self._cache.destroy("timestamps")

        if self._conserve_memory:
            nodecomm = self._comm.Split_type(MPI.COMM_TYPE_SHARED, self._rank)
            nread = nodecomm.size
        else:
            nodecomm = MPI.COMM_SELF
            nread = 1

        self._comm.Barrier()
        timer_tot = Timer()
        timer_tot.start()
        for iread in range(nread):
            timer_step = Timer()
            timer_step.start()
            timer = Timer()
            timer.start()
            if nodecomm.rank % nread == iread:
                self._unstage_signal(detectors, nsamp, obs_period_ranges, signal_type)
            if self._verbose:
                nodecomm.Barrier()
                if self._rank == 0:
                    timer.report_clear(
                        "Unstage signal {} / {}".format(iread + 1, nread)
                    )
            if nodecomm.rank % nread == iread:
                self._unstage_pixels(
                    detectors, nsamp, obs_period_ranges, pixels_dtype, nside
                )
            if self._verbose:
                nodecomm.Barrier()
                if self._rank == 0:
                    timer.report_clear(
                        "Unstage pixels {} / {}".format(iread + 1, nread)
                    )
            if nodecomm.rank % nread == iread:
                self._unstage_pixweights(
                    detectors, nsamp, obs_period_ranges, weight_dtype, nnz, nnz_full
                )
            nodecomm.Barrier()
            if self._verbose and self._rank == 0:
                timer.report_clear(
                    "Unstage pixel weights {} / {}".format(iread + 1, nread)
                )
            if self._rank == 0 and self._verbose and nread > 1:
                timer_step.report_clear("Unstage data {} / {}".format(iread + 1, nread))
        self._comm.Barrier()
        if self._rank == 0 and self._verbose:
            timer_tot.report_clear("Unstage all data")

        del nodecomm
        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import warnings

from ..mpi import use_mpi

import numpy as np

from .. import qarray as qa

from ..op import Operator

from ..timing import function_timer, Timer

conviqt = None

if use_mpi:
    try:
        import libconviqt_wrapper as conviqt
    except ImportError:
        pass


class OpSimConviqt(Operator):
    """Operator which uses libconviqt to generate beam-convolved timestreams.

    This passes through each observation and loops over each detector.
    For each detector, it produces the beam-convolved timestream.

    Args:
        comm (MPI.Comm) : MPI communicator to use for the convolution.
            libConviqt does not work without MPI.
        sky_file (dict or str) : File containing the sky a_lm expansion.
            Tag {detector} will be replaced with the detector name
            If sky_file is a dict, then each detector must have an entry.
        beam_file (dict or str) : File containing the beam a_lm expansion.
            Tag {detector} will be replaced with the detector name.
            If beam_file is a dict, then each detector must have an entry.
        lmax (int) : Maximum ell (and m).  Actual resolution in the
            Healpix FITS file may differ.  If not set, will use the
            maximum expansion order from file.
        beammmax (int) : beam maximum m.  Actual resolution in the
            Healpix FITS file may differ.  If not set, will use the
            maximum expansion order from file.
        pol (bool) : boolean to determine if polarized simulation is needed
        fwhm (float) : width of a symmetric gaussian beam [in arcmin] already
            present in the skyfile (will be deconvolved away).
        order (int) : conviqt order parameter (expert mode)
        calibrate (bool) : Calibrate intensity to 1.0, rather than
            (1 + epsilon) / 2.  Calibrate has no effect if the beam is found
            to be normalized rather than scaled with the leakage factor.
        dxx (bool) : The beam frame is either Dxx or Pxx.  Pxx includes the
            rotation to polarization sensitive basis, Dxx does not.  When
            Dxx=True, detector orientation from attitude quaternions is
            corrected for the polarization angle.
        out (str): the name of the cache object (<name>_<detector>) to
            use for output of the detector timestream.
        mc (int): Monte Carlo index used in synthesizing the input file names.

    """

    def __init__(
        self,
        comm,
        sky_file,
        beam_file,
        lmax=0,
        beammmax=0,
        pol=True,
        fwhm=4.0,
        order=13,
        calibrate=True,
        dxx=True,
        out="conviqt",
        quat_name=None,
        flag_name=None,
        flag_mask=255,
        common_flag_name=None,
        common_flag_mask=255,
        apply_flags=False,
        remove_monopole=False,
        remove_dipole=False,
        normalize_beam=False,
        verbosity=0,
        mc=None,
    ):
        # Call the parent class constructor
        super().__init__()

        self._comm = comm
        self._sky_file = sky_file
        self._beam_file = beam_file
        self._lmax = lmax
        self._beammmax = beammmax
        self._pol = pol
        self._fwhm = fwhm
        self._order = order
        self._calibrate = calibrate
        self._dxx = dxx
        self._quat_name = quat_name
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._apply_flags = apply_flags
        self._remove_monopole = remove_monopole
        self._remove_dipole = remove_dipole
        self._normalize_beam = normalize_beam
        self._verbosity = verbosity
        self._mc = mc

        self._out = out

    @property
    def available(self):
        """Return True if libconviqt is found in the library search path."""
        return conviqt is not None and conviqt.available

    def _check_for_hwp(self, data):
        # Check that there is no rotating HWP (stationary is fine)
        has_hwp = False
        for obs in data.obs:
            try:
                hwpang = obs["tod"].local_hwp_angle()
                if hwpang is not None and np.any(np.abs(hwpang)) > 1e-10:
                    has_hwp = True
                    break
            except:
                pass
        if has_hwp:
            raise RuntimeError(
                "OpSimConviqt cannot handle a HWP.  Please use OpSimWeightedConviqt"
            )
        return

    @function_timer
    def exec(self, data):
        """Loop over all observations and perform the convolution.

        This is done one detector at a time.  For each detector, all data
        products are read from disk.

        Args:
            data (toast.Data): The distributed data.

        """
        if not self.available:
            raise RuntimeError("libconviqt is not available")

        timer = Timer()
        timer.start()

        self._check_for_hwp(data)

        detectors = self._get_detectors(data)

        for det in detectors:
            verbose = self._comm.rank == 0 and self._verbosity > 0

            try:
                sky_file = self._sky_file[det]
            except TypeError:
                sky_file = self._sky_file.format(detector=det, mc=self._mc)
            sky = self.get_sky(sky_file, det, verbose)

            try:
                beam_file = self._beam_file[det]
            except TypeError:
                beam_file = self._beam_file.format(detector=det, mc=self._mc)

            beam = self.get_beam(beam_file, det, verbose)

            detector = self.get_detector(det)

            theta, phi, psi, psi_pol = self.get_pointing(data, det, verbose)
            pnt = self.get_buffer(theta, phi, psi, det, verbose)
            del theta, phi, psi

            convolved_data = self.convolve(sky, beam, detector, pnt, det, verbose)

            self.calibrate(data, det, beam, convolved_data, verbose)
            self.cache(data, det, convolved_data, verbose)

            del pnt, detector, beam, sky

            if verbose:
                timer.report_clear("conviqt process detector {}".format(det))

        return

    def _get_detectors(self, data):
        """Assemble a list of detectors across all processes and
        observations in `self._comm`.
        """
        dets = set()
        for obs in data.obs:
            tod = obs["tod"]
            for det in tod.local_dets:
                dets.add(det)
        all_dets = self._comm.gather(dets, root=0)
        if self._comm.rank == 0:
            for some_dets in all_dets:
                dets.update(some_dets)
            dets = sorted(dets)
        all_dets = self._comm.bcast(dets, root=0)
        return all_dets

    def _get_psi_pol(self, focalplane, det):
        """Parse polarization angle in radians from the focalplane
        dictionary.  The angle is relative to the Pxx basis.
        """
        if det not in focalplane:
            raise RuntimeError("focalplane does not include {}".format(det))
        if "psi_pol_deg" in focalplane[det]:
            psi_pol = np.radians(focalplane[det]["psi_pol_deg"])
        elif "psi_pol_rad" in focalplane[det]:
            psi_pol = focalplane[det]["psi_pol_rad"]
        elif "pol_angle_deg" in focalplane[det]:
            warnings.warn(
                "Use psi_pol and psi_uv rather than pol_angle", DeprecationWarning
            )
            psi_pol = np.radians(focalplane[det]["pol_angle_deg"])
        elif "pol_angle_rad" in focalplane[det]:
            warnings.warn(
                "Use psi_pol and psi_uv rather than pol_angle", DeprecationWarning
            )
            psi_pol = focalplane[det]["pol_angle_rad"]
        else:
            raise RuntimeError("focalplane[{}] does not include psi_pol".format(det))
        return psi_pol

    def _get_psi_uv(self, focalplane, det):
        """Parse Pxx basis angle in radians from the focalplane
        dictionary.  The angle is measured from Dxx to Pxx basis.
        """
        if det not in focalplane:
            raise RuntimeError("focalplane does not include {}".format(det))
        if "psi_uv_deg" in focalplane[det]:
            psi_uv = np.radians(focalplane[det]["psi_uv_deg"])
        elif "psi_uv_rad" in focalplane[det]:
            psi_uv = focalplane[det]["psi_uv_rad"]
        else:
            raise RuntimeError("focalplane[{}] does not include psi_uv".format(det))
        return psipol

    def _get_epsilon(self, focalplane, det):
        """Parse polarization leakage (epsilon) from the focalplane
        object or dictionary.
        """
        if det not in focalplane:
            raise RuntimeError("focalplane does not include {}".format(det))
        if "pol_leakage" in focalplane[det]:
            epsilon = focalplane[det]["pol_leakage"]
        else:
            # Assume zero polarization leakage
            epsilon = 0
        return epsilon

    def get_sky(self, skyfile, det, verbose):
        timer = Timer()
        timer.start()
        sky = conviqt.Sky(self._lmax, self._pol, skyfile, self._fwhm, self._comm)
        if self._remove_monopole:
            sky.remove_monopole()
        if self._remove_dipole:
            sky.remove_dipole()
        if verbose:
            timer.report_clear("initialize sky for detector {}".format(det))
        return sky

    def get_beam(self, beamfile, det, verbose):
        timer = Timer()
        timer.start()
        beam = conviqt.Beam(self._lmax, self._beammmax, self._pol, beamfile, self._comm)
        if self._normalize_beam:
            beam.normalize()
        if verbose:
            timer.report_clear("initialize beam for detector {}".format(det))
        return beam

    def get_detector(self, det):
        """We always create the detector with zero leakage and scale
        the returned TOD ourselves
        """
        detector = conviqt.Detector(name=det, epsilon=0)
        return detector

    def get_pointing(self, data, det, verbose):
        """Return the detector pointing as ZYZ Euler angles without the
        polarization sensitive angle.  These angles are to be compatible
        with Pxx or Dxx frame beam products
        """
        # We need the three pointing angles to describe the
        # pointing.  local_pointing() returns the attitude quaternions.
        nullquat = np.array([0, 0, 0, 1], dtype=np.float64)
        timer = Timer()
        timer.start()
        all_theta, all_phi, all_psi, all_psi_pol = [], [], [], []
        for obs in data.obs:
            tod = obs["tod"]
            if det not in tod.local_dets:
                continue
            focalplane = obs["focalplane"]
            quats = tod.local_pointing(det, self._quat_name)
            if verbose:
                timer.report_clear("get detector pointing for {}".format(det))

            if self._apply_flags:
                common = tod.local_common_flags(self._common_flag_name)
                flags = tod.local_flags(det, self._flag_name)
                common = common & self._common_flag_mask
                flags = flags & self._flag_mask
                totflags = np.copy(flags)
                totflags |= common
                quats = quats.copy()
                quats[totflags != 0] = nullquat
                if verbose:
                    timer.report_clear("initialize flags for detector {}".format(det))

            theta, phi, psi = qa.to_angles(quats)
            # Polarization angle in the Pxx basis
            psi_pol = self._get_psi_pol(focalplane, det)
            if self._dxx:
                # Add angle between Dxx and Pxx
                psi_pol += self._get_psi_uv(focalplane, det)
            psi -= psi_pol
            # Add also a potential HWP angle
            psi_pol = np.ones(psi.size) * psi_pol
            try:
                hwpang = tod.local_hwp_angle()
                psi_pol += 2 * hwpang
            except:
                pass
            all_theta.append(theta)
            all_phi.append(phi)
            all_psi.append(psi)
            all_psi_pol.append(psi_pol)
        if len(all_theta) > 0:
            all_theta = np.hstack(all_theta)
            all_phi = np.hstack(all_phi)
            all_psi = np.hstack(all_psi)
            all_psi_pol = np.hstack(all_psi_pol)
        if verbose:
            timer.report_clear("compute pointing angles for detector {}".format(det))
        return all_theta, all_phi, all_psi, all_psi_pol

    def get_buffer(self, theta, phi, psi, det, verbose):
        """Pack the pointing into the conviqt pointing array"""
        timer = Timer()
        timer.start()
        pnt = conviqt.Pointing(len(theta))
        if pnt._nrow > 0:
            arr = pnt.data()
            arr[:, 0] = phi
            arr[:, 1] = theta
            arr[:, 2] = psi
        if verbose:
            timer.report_clear("pack input array for detector {}".format(det))
        return pnt

    def convolve(self, sky, beam, detector, pnt, det, verbose):
        timer = Timer()
        timer.start()
        convolver = conviqt.Convolver(
            sky,
            beam,
            detector,
            self._pol,
            self._lmax,
            self._beammmax,
            self._order,
            self._verbosity,
            self._comm,
        )
        convolver.convolve(pnt)
        if verbose:
            timer.report_clear("convolve detector {}".format(det))

        # The pointer to the data will have changed during
        # the convolution call ...

        if pnt._nrow > 0:
            arr = pnt.data()
            convolved_data = arr[:, 3].astype(np.float64)
        else:
            convolved_data = None
        if verbose:
            timer.report_clear("extract convolved data for {}".format(det))

        del convolver

        return convolved_data

    def calibrate(self, data, det, beam, convolved_data, verbose):
        """By default, libConviqt results returns a signal that conforms to
        TOD = (1 + epsilon) / 2 * intensity + (1 - epsilon) / 2 * polarization.

        When calibrate = True, we rescale the TOD to
        TOD = intensity + (1 - epsilon) / (1 + epsilon) * polarization
        """
        if not self._calibrate or beam.normalized():
            return
        timer = Timer()
        timer.start()
        offset = 0
        for obs in data.obs:
            tod = obs["tod"]
            if det not in tod.local_dets:
                continue
            focalplane = obs["focalplane"]
            epsilon = self._get_epsilon(focalplane, det)
            nsample = tod.local_samples[1]
            convolved_data[offset : offset + nsample] *= 2 / (1 + epsilon)
            offset += nsample
        if verbose:
            timer.report_clear("calibrate detector {}".format(det))
        return

    def cache(self, data, det, convolved_data, verbose):
        """Inject the convolved data into the TOD cache."""
        timer = Timer()
        timer.start()
        offset = 0
        for obs in data.obs:
            tod = obs["tod"]
            if det not in tod.local_dets:
                continue
            nsample = tod.local_samples[1]
            cachename = "{}_{}".format(self._out, det)
            if not tod.cache.exists(cachename):
                tod.cache.create(cachename, np.float64, (nsample,))
            ref = tod.cache.reference(cachename)
            ref[:] += convolved_data[offset : offset + nsample]
            offset += nsample
        if verbose:
            timer.report_clear("cache detector {}".format(det))
        return


class OpSimWeightedConviqt(OpSimConviqt):
    """Operator which uses libconviqt to generate beam-convolved timestreams.
    This operator should be used in presence of a spinning  HWP which  makes the beam time-dependent,
    constantly mapping the co- and cross polar responses on to each other.  In OpSimConviqt we assume the  beam to be static.

    This passes through each observation and loops over each detector.
    For each detector, it produces the beam-convolved timestream.

    Args:
        comm (MPI.Comm) : MPI communicator to use for the convolution.
            libConviqt does not work without MPI.
        sky_file (dict or str) : File containing the sky a_lm expansion.
            Tag {detector} will be replaced with the detector name
            If sky_file is a dict, then each detector must have an entry.
        beam_file (dict or str) : File containing the beam a_lm expansion.
            Tag {detector} will be replaced with the detector name.
            If beam_file is a dict, then each detector must have an entry.
        lmax (int) : Maximum ell (and m).  Actual resolution in the
            Healpix FITS file may differ.  If not set, will use the
            maximum expansion order from file.
        beammmax (int) : beam maximum m.  Actual resolution in the
            Healpix FITS file may differ.  If not set, will use the
            maximum expansion order from file.
        pol (bool) : boolean to determine if polarized simulation is needed
        fwhm (float) : width of a symmetric gaussian beam [in arcmin] already
            present in the skyfile (will be deconvolved away).
        order (int) : conviqt order parameter (expert mode)
        calibrate (bool) : Calibrate intensity to 1.0, rather than
            (1 + epsilon) / 2.  Calibrate has no effect if the beam is found
            to be normalized rather than scaled with the leakage factor.
        dxx (bool) : The beam frame is either Dxx or Pxx.  Pxx includes the
            rotation to polarization sensitive basis, Dxx does not.  When
            Dxx=True, detector orientation from attitude quaternions is
            corrected for the polarization angle.
        out (str): the name of the cache object (<name>_<detector>) to
            use for output of the detector timestream.
        mc (int): Monte Carlo index used in synthesizing the input file names.

    """

    def __init__(
        self,
        comm,
        sky_file,
        beam_file,
        lmax=0,
        beammmax=0,
        pol=True,
        fwhm=4.0,
        order=13,
        calibrate=True,
        dxx=True,
        out="conviqt",
        quat_name=None,
        flag_name=None,
        flag_mask=255,
        common_flag_name=None,
        common_flag_mask=255,
        apply_flags=False,
        remove_monopole=False,
        remove_dipole=False,
        normalize_beam=False,
        verbosity=0,
        mc=None,
    ):
        # Call the parent class constructor
        super(OpSimConviqt, self).__init__()

        self._comm = comm
        self._sky_file = sky_file
        self._beam_file = beam_file
        self._lmax = lmax
        self._beammmax = beammmax
        self._pol = pol
        self._fwhm = fwhm
        self._order = order
        self._calibrate = calibrate
        self._dxx = dxx
        self._quat_name = quat_name
        self._flag_name = flag_name
        self._flag_mask = flag_mask
        self._common_flag_name = common_flag_name
        self._common_flag_mask = common_flag_mask
        self._apply_flags = apply_flags
        self._remove_monopole = remove_monopole
        self._remove_dipole = remove_dipole
        self._normalize_beam = normalize_beam
        self._verbosity = verbosity
        self._mc = mc

        self._out = out

    @function_timer
    def exec(self, data):
        """Loop over all observations and perform the convolution.

        This is done one detector at a time.  For each detector, all data
        products are read from disk.

        Args:
            data (toast.Data): The distributed data.

        """
        if not self.available:
            raise RuntimeError("libconviqt is not available")

        timer = Timer()
        timer.start()

        detectors = self._get_detectors(data)

        for det in detectors:
            verbose = self._comm.rank == 0 and self._verbosity > 0

            try:
                sky_file = self._sky_file[det]
            except TypeError:
                sky_file = self._sky_file.format(detector=det, mc=self._mc)
            sky = self.get_sky(sky_file, det, verbose)

            try:
                beam_file = self._beam_file[det]
            except TypeError:
                beam_file = self._beam_file.format(detector=det, mc=self._mc)
                beam_file_i00 = beam_file.replace(".fits", "_I000.fits")
                beam_file_0i0 = beam_file.replace(".fits", "_0I00.fits")
                beam_file_00i = beam_file.replace(".fits", "_00I0.fits")

            beamI00 = self.get_beam(beam_file_i00, det, verbose)
            beam0I0 = self.get_beam(beam_file_0i0, det, verbose)
            beam00I = self.get_beam(beam_file_00i, det, verbose)
            detector = self.get_detector(det)

            theta, phi, psi, psi_pol = self.get_pointing(data, det, verbose)

            # I-beam convolution
            pnt = self.get_buffer(theta, phi, psi, det, verbose)
            convolved_data = self.convolve(sky, beamI00, detector, pnt, det, verbose)

            # Q-beam convolution
            del pnt
            pnt = self.get_buffer(theta, phi, psi, det, verbose)
            convolved_data += np.cos(2 * psi_pol) * self.convolve(
                sky, beam0I0, detector, pnt, det, verbose
            )

            # U-beam convolution
            del pnt
            pnt = self.get_buffer(theta, phi, psi, det, verbose)
            convolved_data += np.sin(2 * psi_pol) * self.convolve(
                sky, beam00I, detector, pnt, det, verbose
            )
            del theta, phi, psi

            self.calibrate(data, det, beamI00, convolved_data, verbose)

            self.cache(data, det, convolved_data, verbose)

            del pnt, detector, beamI00, beam0I0, beam00I, sky

            if verbose:
                timer.report_clear("conviqt process detector {}".format(det))

        return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import numpy as np

from ._libtoast import FFTPlanType, FFTDirection, FFTPlanReal1D, FFTPlanReal1DStore

from .utils import object_ndim


def r1d_forward(indata):
    """High level forward FFT interface to internal library.

    This function uses the internal store of FFT plans to do a forward
    1D real FFT.  Data is copied into the internal aligned memory
    buffer inside the plan and the result is copied out and returned.

    If a 2D array is passed, the first dimension is assumed to be the number of
    FFTs to batch at once.

    Args:
        indata (array): The input data array.

    Returns:
        (array): The output Fourier-domain data in FFTW half-complex format.

    """
    ndim = object_ndim(indata)
    count = None
    length = None
    if ndim == 1:
        # Just a single FFT
        count = 1
        length = len(indata)
    else:
        # We must have a batch of FFTs
        count = indata.shape[0]
        length = indata.shape[1]

    store = FFTPlanReal1DStore.get()
    plan = store.forward(length, count)

    if count == 1:
        plan.tdata(0)[:] = indata
    else:
        for indx in range(count):
            plan.tdata(indx)[:] = indata[indx]
    plan.exec()
    if count == 1:
        return np.array(plan.fdata(0))
    else:
        ret = np.zeros_like(indata)
        for indx in range(count):
            ret[indx, :] = plan.fdata(indx)
        return ret


def r1d_backward(indata):
    """High level backward FFT interface to internal library.

    This function uses the internal store of FFT plans to do a backward
    1D real FFT.  Data is copied into the internal aligned memory
    buffer inside the plan and the result is copied out and returned.

    If a 2D array is passed, the first dimension is assumed to be the number of
    FFTs to batch at once.

    Args:
        indata (array): The input Fourier-domain data array in FFTW
            half-complex format.

    Returns:
        (array): The output time domain data.

    """
    ndim = object_ndim(indata)
    count = None
    length = None
    if ndim == 1:
        # Just a single FFT
        count = 1
        length = len(indata)
    else:
        # We must have a batch of FFTs
        count = indata.shape[0]
        length = indata.shape[1]

    store = FFTPlanReal1DStore.get()
    plan = store.backward(length, count)

    if count == 1:
        plan.fdata(0)[:] = indata
    else:
        for indx in range(count):
            plan.fdata(indx)[:] = indata[indx]
    plan.exec()
    if count == 1:
        return np.array(plan.tdata(0))
    else:
        ret = np.zeros_like(indata)
        for indx in range(count):
            ret[indx, :] = plan.tdata(indx)
        return ret



#
#  Time Ordered Astrophysics Scalable Tools (TOAST)
#
# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.
#
"""Time Ordered Astrophysics Scalable Tools (TOAST) is a software package
designed to allow the processing of data from telescopes that acquire
data as timestreams (rather than images).
"""
import sys
import os

# Get the package version from the libtoast environment if possible.  If this
# import fails, it is likely due to the toast package being imported prior to
# the build by setuptools (for example).
__version__ = None
try:
    from ._libtoast import Environment

    env = Environment.get()
    __version__ = env.version()
except ImportError:
    # import traceback
    # exc_type, exc_value, exc_traceback = sys.exc_info()
    # lines = traceback.format_exception(exc_type, exc_value, exc_traceback)
    # print("".join(lines), flush=True)
    #
    # Just manually read the release file.
    thisdir = os.path.abspath(os.path.dirname(__file__))
    relfile = os.path.join(thisdir, "RELEASE")
    try:
        with open(relfile, "r") as rel:
            if __version__ is None:
                __version__ = rel.readline().rstrip()
    except:
        raise ImportError("Cannot read RELEASE file")

# Namespace imports
from .mpi import Comm

from .dist import Data, distribute_uniform, distribute_discrete, distribute_samples

from .op import Operator

from .weather import Weather



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# This file provides a simplified interface to quaternion operations.

import numpy as np

from .utils import Logger, AlignedF64, ensure_buffer_f64, object_ndim

from ._libtoast import (
    qa_inv,
    qa_amplitude,
    qa_normalize,
    qa_rotate,
    qa_mult,
    qa_slerp,
    qa_exp,
    qa_ln,
    qa_pow,
    qa_from_axisangle,
    qa_to_axisangle,
    qa_to_rotmat,
    qa_from_rotmat,
    qa_from_vectors,
    qa_from_angles,
    qa_to_angles,
    qa_to_position,
    qa_from_position,
)


def inv(q):
    """Invert a quaternion array.

    Args:
        q (array_like):  The quaternion array to invert.

    Returns:
        (array):  The inverse.

    """
    qin = ensure_buffer_f64(q)
    out = AlignedF64(len(qin))
    out[:] = qin
    qa_inv(out)
    if len(out) == 4:
        if object_ndim(q) == 2:
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def amplitude(q):
    """Amplitude of a quaternion array

    Args:
        q (array_like):  The quaternion array.

    Returns:
        (array):  The array of amplitudes.

    """
    qin = ensure_buffer_f64(q)
    lamp = len(qin) // 4
    amp = AlignedF64(lamp)
    qa_amplitude(qin, amp)
    if len(amp) == 1:
        if object_ndim(q) == 2:
            return amp.array()
        else:
            return float(amp[0])
    else:
        return amp.array()


def norm(q):
    """Normalize quaternion array.

    Args:
        q (array_like):  The quaternion array.

    Returns:
        (array):  The normalized array.

    """
    qin = ensure_buffer_f64(q)
    lq = len(qin)
    nrm = AlignedF64(lq)
    qa_normalize(qin, nrm)
    if len(qin) == 4:
        if object_ndim(q) == 2:
            return nrm.array().reshape((1, 4))
        else:
            return nrm.array()
    else:
        return nrm.array().reshape((-1, 4))


def rotate(q, v):
    """Rotate vectors with quaternions.

    The number of quaternions and vectors should either be equal or one of
    the arrays should be a single quaternion or vector.

    Args:
        q (array_like):  The quaternion array.
        v (array_like):  The vector array.

    Returns:
        (array):  The rotated vectors.

    """
    qin = ensure_buffer_f64(q)
    nq = len(qin) // 4
    vin = ensure_buffer_f64(v)
    nv = len(vin) // 3
    nout = None
    if nq > nv:
        nout = nq
    else:
        nout = nv
    out = AlignedF64(3 * nout)
    qa_rotate(qin, vin, out)
    if len(out) == 3:
        if (object_ndim(q) == 2) or (object_ndim(v) == 2):
            return out.array().reshape(1, 3)
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 3))


def mult(p, q):
    """Multiply arrays of quaternions.

    The number of quaternions in the input arrays should either be equal or
    one.

    Args:
        p (array_like):  The first quaternion array.
        q (array_like):  The second quaternion array.

    Returns:
        (array):  The product.

    """
    pin = ensure_buffer_f64(p)
    qin = ensure_buffer_f64(q)
    out = None
    if len(pin) > len(qin):
        out = AlignedF64(len(pin))
    else:
        out = AlignedF64(len(qin))
    qa_mult(pin, qin, out)
    if len(out) == 4:
        if (object_ndim(p) == 2) or (object_ndim(q) == 2):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def slerp(targettime, time, q):
    """Spherical Linear Interpolation (SLERP) of a quaternion array.

    The input quaternions are specified at time stamps, and the output is
    interpolated to the target times.

    Args:
        targettime (array_like):  The output target times.
        time (array_like):  The input times.
        q (array_like):  The quaternion array.

    Returns:
        (array):  The interpolated quaternions.

    """
    tgt = ensure_buffer_f64(targettime)
    t = ensure_buffer_f64(time)
    qin = ensure_buffer_f64(q)
    log = Logger.get()
    if len(t) < 2:
        msg = "SLERP input times must have at least two values"
        log.error(msg)
        raise RuntimeError(msg)
    out = AlignedF64(4 * len(tgt))
    qa_slerp(t, tgt, qin, out)
    if len(out) == 4:
        if object_ndim(targettime) == 1:
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def exp(q):
    """Exponential of a quaternion array.

    Args:
        q (array_like):  The quaternion array.

    Returns:
        (array):  The result.

    """
    qin = ensure_buffer_f64(q)
    out = AlignedF64(len(qin))
    qa_exp(qin, out)
    if len(out) == 4:
        if object_ndim(q) == 2:
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def ln(q):
    """Natural logarithm of a quaternion array.

    Args:
        q (array_like):  The quaternion array.

    Returns:
        (array):  The result.

    """
    qin = ensure_buffer_f64(q)
    out = AlignedF64(len(qin))
    qa_ln(qin, out)
    if len(out) == 4:
        if object_ndim(q) == 2:
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def pow(q, pw):
    """Real power of a quaternion array.

    Args:
        q (array_like):  The quaternion array.
        pw (array_like):  The power.

    Returns:
        (array):  The result.

    """
    qin = ensure_buffer_f64(q)
    pwin = ensure_buffer_f64(pw)
    out = AlignedF64(len(qin))
    qa_pow(qin, pwin, out)
    if len(out) == 4:
        if (object_ndim(q) == 2) or (object_ndim(pw) == 1):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def rotation(axis, angle):
    """Create quaternions from axis / angle information.

    Args:
        axis (array_like):  The array of normalized axis vectors.
        angle (array_like):  The array of angles (in radians).

    Returns:
        (array):  The result.

    """
    if np.shape(axis)[-1] != 3:
        raise RuntimeError("axis is not a 3D vector")
    axin = ensure_buffer_f64(axis)
    angin = ensure_buffer_f64(angle)
    out = AlignedF64(4 * len(angin))
    qa_from_axisangle(axin, angin, out)
    if len(out) == 4:
        if (object_ndim(axis) == 2) or (object_ndim(angle) == 1):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def to_axisangle(q):
    """Convert quaterions to axis / angle form.

    Args:
        q (array_like):  The input quaternions.

    Returns:
        (tuple):  The (axis, angle) results.

    """
    qin = ensure_buffer_f64(q)
    lq = len(qin) // 4
    ax = AlignedF64(3 * lq)
    ang = AlignedF64(lq)
    qa_to_axisangle(qin, ax, ang)
    if len(ax) == 3:
        if object_ndim(q) == 2:
            return (ax.array().reshape((1, 3)), ang.array())
        else:
            return (ax.array(), float(ang[0]))
    else:
        return (ax.array().reshape((-1, 3)), ang.array())


def to_rotmat(q):
    """Convert quaternions to rotation matrices.

    Args:
        q (array_like):  The input quaternions.

    Returns:
        (array):  The rotation matrices.

    """
    qin = ensure_buffer_f64(q)
    lq = len(qin) // 4
    out = AlignedF64(9 * lq)
    qa_to_rotmat(qin, out)
    if len(out) == 9:
        if object_ndim(q) == 2:
            return out.array().reshape((1, 3, 3))
        else:
            return out.array().reshape((3, 3))
    else:
        return out.array().reshape((-1, 3, 3))


def from_rotmat(rotmat):
    """Create quaternions from rotation matrices.

    Args:
        rotmat (array_like):  The input 3x3 rotation matrices.

    Returns:
        (array):  The quaternions.

    """
    rot = ensure_buffer_f64(rotmat)
    lr = len(rot) // 9
    out = AlignedF64(4 * lr)
    qa_from_rotmat(rot, out)
    if len(out) == 4:
        if object_ndim(rotmat) == 3:
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def from_vectors(v1, v2):
    """Create quaternions from pairs of vectors.

    Args:
        v1 (array_like):  The input starting vectors.
        v2 (array_like):  The input ending vectors.

    Returns:
        (array):  The quaternions.

    """
    v1in = ensure_buffer_f64(v1)
    v2in = ensure_buffer_f64(v2)
    lv = len(v1in) // 3
    out = AlignedF64(4 * lv)
    qa_from_vectors(v1in, v2in, out)
    if len(out) == 4:
        if (object_ndim(v1) == 2) or (object_ndim(v2) == 2):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def from_angles(theta, phi, pa, IAU=False):
    """Create quaternions from spherical coordinates.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.  The position angle is with respect
    to the local meridian at the point described by the theta / phi
    coordinates.

    Args:
        theta (array_like):  The input theta angles.
        phi (array_like):  The input phi vectors.
        pa (array_like):  The input position angle vectors.
        IAU (bool):  If True, use IAU convention.

    Returns:
        (array):  The quaternions.

    """
    thetain = ensure_buffer_f64(theta)
    phiin = ensure_buffer_f64(phi)
    pain = ensure_buffer_f64(pa)
    lt = len(thetain)
    out = AlignedF64(4 * lt)
    qa_from_angles(thetain, phiin, pain, out, IAU)
    if len(out) == 4:
        if (
            (object_ndim(theta) == 1)
            or (object_ndim(phi) == 1)
            or (object_ndim(pa) == 1)
        ):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def to_angles(q, IAU=False):
    """Convert quaternions to spherical coordinates and position angle.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.  The position angle is with respect
    to the local meridian at the point described by the theta / phi
    coordinates.

    Args:
        q (array_like):  The input quaternions.
        IAU (bool):  If True, use IAU convention.

    Returns:
        (tuple):  The (theta, phi, pa) arrays.

    """
    qin = ensure_buffer_f64(q)
    lq = len(qin) // 4
    theta = AlignedF64(lq)
    phi = AlignedF64(lq)
    pa = AlignedF64(lq)
    qa_to_angles(qin, theta, phi, pa, IAU)
    if len(qin) == 4:
        if object_ndim(q) == 2:
            return (theta.array(), phi.array(), pa.array())
        else:
            return (float(theta[0]), float(phi[0]), float(pa[0]))
    return (theta.array(), phi.array(), pa.array())


def from_position(theta, phi):
    """Create quaternions from spherical coordinates.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.

    Args:
        theta (array_like):  The input theta angles.
        phi (array_like):  The input phi vectors.

    Returns:
        (array):  The quaternions.

    """
    thetain = ensure_buffer_f64(theta)
    phiin = ensure_buffer_f64(phi)
    lt = len(thetain)
    out = AlignedF64(4 * lt)
    qa_from_position(thetain, phiin, out)
    if len(out) == 4:
        if (object_ndim(theta) == 1) or (object_ndim(phi) == 1):
            return out.array().reshape((1, 4))
        else:
            return out.array()
    else:
        return out.array().reshape((-1, 4))


def to_position(q):
    """Convert quaternions to spherical coordinates.

    The theta angle is measured down from the North pole and phi is
    measured from the prime meridian.

    Args:
        q (array_like):  The input quaternions.

    Returns:
        (tuple):  The (theta, phi) arrays.

    """
    qin = ensure_buffer_f64(q)
    lq = len(qin) // 4
    theta = AlignedF64(lq)
    phi = AlignedF64(lq)
    qa_to_position(qin, theta, phi)
    if len(qin) == 4:
        if object_ndim(q) == 2:
            return (theta.array(), phi.array())
        else:
            return (float(theta[0]), float(phi[0]))
    return (theta.array(), phi.array())



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import time

import numpy as np
import numpy.testing as nt

from .mpi import MPITestCase

from ..utils import Environment

from ..mpi import MPIShared, MPILock

from ._helpers import create_comm


class EnvTest(MPITestCase):
    def setUp(self):
        self.rank = 0
        self.nproc = 1
        if self.comm is not None:
            self.rank = self.comm.rank
            self.nproc = self.comm.size

    def test_env(self):
        env = Environment.get()
        if self.rank == 0:
            print(env, flush=True)

    def test_comm(self):
        comm = create_comm(self.comm)
        for p in range(self.nproc):
            if p == self.rank:
                print(comm, flush=True)
            if self.comm is not None:
                self.comm.barrier()

    def test_mpi_shared(self):
        """This is based on the simple tests from the upstream repo:
        https://github.com/tskisner/mpi_shmem/blob/master/test.py
        """
        # Dimensions of our shared memory array
        datadims = (2, 5, 10)

        # Dimensions of the incremental slab that we will
        # copy during each set() call.
        updatedims = (1, 1, 5)

        # How many updates are there to cover the whole
        # data array?
        nupdate = 1
        for d in range(len(datadims)):
            nupdate *= datadims[d] // updatedims[d]

        for datatype in [np.float64, np.float32, np.int64, np.int32]:
            # For testing the "set()" method, every process is going to
            # create a full-sized data buffer and fill it with its process rank.
            local = np.ones(datadims, dtype=datatype)
            local *= self.rank

            with MPIShared(local.shape, local.dtype, self.comm) as shm:
                for p in range(self.nproc):
                    # Every process takes turns writing to the buffer.
                    setdata = None
                    setoffset = (0, 0, 0)

                    # Write to the whole data volume, but in small blocks
                    for upd in range(nupdate):
                        if p == self.rank:
                            # My turn!  Write my process rank to the buffer slab.
                            setdata = local[
                                setoffset[0] : setoffset[0] + updatedims[0],
                                setoffset[1] : setoffset[1] + updatedims[1],
                                setoffset[2] : setoffset[2] + updatedims[2],
                            ]
                        try:
                            # All processes call set(), but only data on rank p matters.
                            shm.set(setdata, setoffset, fromrank=p)
                        except:
                            print(
                                "proc {} threw exception during set()".format(
                                    self.rank
                                ),
                                flush=True,
                            )
                            raise RuntimeError("shared memory set failed")

                        # Increment the write offset within the array

                        x = setoffset[0]
                        y = setoffset[1]
                        z = setoffset[2]

                        z += updatedims[2]
                        if z >= datadims[2]:
                            z = 0
                            y += updatedims[1]
                        if y >= datadims[1]:
                            y = 0
                            x += updatedims[0]

                        setoffset = (x, y, z)

                    # Every process is now going to read a copy from the shared memory
                    # and make sure that they see the data written by the current process.
                    check = np.zeros_like(local)
                    check[:, :, :] = shm[:, :, :]

                    truth = np.ones_like(local)
                    truth *= p

                    # This should be bitwise identical, even for floats
                    nt.assert_equal(check[:, :, :], truth[:, :, :])

                # Ensure that we can reference the memory buffer from numpy without
                # a memory copy.  The intention is that a slice of the shared memory
                # buffer should appear as a C-contiguous ndarray whenever we slice
                # along the last dimension.

                for p in range(self.nproc):
                    if p == self.rank:
                        slc = shm[1, 2]
                        print(
                            "proc {} slice has dims {}, dtype {}, C = {}".format(
                                p, slc.shape, slc.dtype.str, slc.flags["C_CONTIGUOUS"]
                            ),
                            flush=True,
                        )
                    if self.comm is not None:
                        self.comm.barrier()

    def test_mpi_lock(self):
        """This is based on the simple tests from the upstream repo:
        https://github.com/tskisner/mpi_shmem/blob/master/test.py
        """
        sleepsec = 0.1
        lock = MPILock(self.comm, root=0, debug=True)

        msg = "test lock:  process {} got the lock".format(self.rank)

        lock.lock()
        print(msg, flush=True)
        time.sleep(sleepsec)
        lock.unlock()

        if self.comm is not None:
            self.comm.barrier()



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

import inspect

from functools import wraps

import csv

from collections import OrderedDict

import numpy as np

from ._libtoast import Timer, GlobalTimers

from .utils import Environment


def function_timer(f):
    env = Environment.get()
    ft = env.function_timers()
    if ft:
        nm = None
        if inspect.ismethod(f):
            nm = f.__self__.__name__
        else:
            nm = f.__qualname__
        tnm = "{} (function_timer)".format(nm)

        @wraps(f)
        def df(*args, **kwargs):
            gt = GlobalTimers.get()
            gt.start(tnm)
            result = f(*args, **kwargs)
            gt.stop(tnm)
            return result

    else:

        @wraps(f)
        def df(*args, **kwargs):
            return f(*args, **kwargs)

    return df


def compute_stats(plist, full=False):
    """Compute the global timer properties.

    Given a list of dictionaries (one per process), examine the timers in each
    and compute the min / max / mean of the values and number of calls.

    Args:
        plist (list):  The list of per-process results.
        full (bool):  If True, return the full set of process values for each counter.

    Returns:
        (dict):  The stats for each timer.

    """
    # First build the set of all timers
    nproc = len(plist)
    allnames = set()
    for p in plist:
        for k in p.keys():
            allnames.add(k)
    # Repack the timer values and call count into arrays:
    seconds = dict()
    calls = dict()
    for nm in allnames:
        seconds[nm] = np.empty(nproc, dtype=np.float64)
        calls[nm] = np.empty(nproc, dtype=np.int64)
    idx = 0
    for p in plist:
        for nm in allnames:
            if nm in p:
                seconds[nm][idx] = p[nm].seconds()
                calls[nm][idx] = p[nm].calls()
            else:
                seconds[nm][idx] = -1.0
                calls[nm][idx] = -1
        idx += 1
    result = dict()
    for nm in allnames:
        result[nm] = dict()
        good = calls[nm] >= 0
        result[nm]["participating"] = np.sum(good)
        result[nm]["call_min"] = np.min(calls[nm][good])
        result[nm]["call_max"] = np.max(calls[nm][good])
        result[nm]["call_mean"] = np.mean(calls[nm][good])
        result[nm]["call_median"] = np.median(calls[nm][good])
        result[nm]["time_min"] = np.min(seconds[nm][good])
        result[nm]["time_max"] = np.max(seconds[nm][good])
        result[nm]["time_mean"] = np.mean(seconds[nm][good])
        result[nm]["time_median"] = np.median(seconds[nm][good])
        if full:
            result[nm]["calls"] = calls[nm]
            result[nm]["times"] = seconds[nm]
    return result


def gather_timers(comm=None, root=0):
    """Gather global timer information from across a communicator.

    Args:
        comm (MPI.Comm):  The communicator or None.
        root (int):  The process returning the results.

    Returns:
        (dict):  The timer stats on the root process, otherwise None.

    """
    gt = GlobalTimers.get()
    local = gt.collect()
    all = None
    if comm is None:
        all = [local]
    else:
        all = comm.gather(local, root=root)
    result = None
    if (comm is None) or (comm.rank == root):
        result = compute_stats(all)
    return result


def dump(results, path):
    """Write out timing results to a format suitable for further processing.

    Args:
        results (dict):  The results as returned by compute_stats().
        path (str):  File root name to dump.

    Returns:
        None

    """
    cols = OrderedDict(
        [
            ("Timer", "Name"),
            ("Processes", "participating"),
            ("Minimum Calls", "call_min"),
            ("Maximum Calls", "call_max"),
            ("Minimum Time", "time_min"),
            ("Maximum Time", "time_max"),
            ("Mean Time", "time_mean"),
            ("Median Time", "time_median"),
        ]
    )
    outpath = "{}.csv".format(path)
    with open(outpath, "w", newline="") as f:
        w = csv.writer(f, delimiter=",", quotechar="'")
        w.writerow(cols.keys())
        for nm, props in results.items():
            row = [nm]
            for k, v in cols.items():
                if k == "Timer":
                    continue
                row.append(props[v])
            w.writerow(row)
    return



# Copyright (c) 2015-2020 by the parties listed in the AUTHORS file.
# All rights reserved.  Use of this source code is governed by
# a BSD-style license that can be found in the LICENSE file.

# import functions in our public API

from .pixels import DistPixels

from .cov import (
    covariance_invert,
    covariance_rcond,
    covariance_multiply,
    covariance_apply,
)
